# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')

import sys
import os
import io
import errno
import mmap
import struct
from threading import Thread, Event

import numpy as np


GAMMA_DEFAULT = [
    0,  0,  0,  0,  0,  0,  1,  1,
    2,  2,  3,  3,  4,  5,  6,  7,
    8,  9,  10, 11, 12, 14, 15, 17,
    18, 20, 21, 23, 25, 27, 29, 31]
GAMMA_LOW = [
    0,  1,  1,  1,  1,  1,  1,  1,
    1,  1,  1,  1,  1,  2,  2,  2,
    3,  3,  3,  4,  4,  5,  5,  6,
    6,  7,  7,  8,  8,  9,  10, 10]


def screen_filename():
    # returns the file used to represent the state of the emulated sense HAT's
    # screen. On UNIX we try /dev/shm then fall back to /tmp; on Windows we
    # use whatever %TEMP% points at
    fname = 'rpi-sense-emu-screen'
    if sys.platform.startswith('win'):
        # just use a temporary file on Windows
        return os.path.join(os.environ['TEMP'], fname)
    else:
        if os.path.exists('/dev/shm'):
            return os.path.join('/dev/shm', fname)
        else:
            return os.path.join('/tmp', fname)


def init_screen():
    try:
        # Attempt to open the screen's device file and ensure it's 160 bytes
        # long
        fd = io.open(screen_filename(), 'r+b', buffering=0)
        fd.seek(160)
        fd.truncate()
    except IOError as e:
        # If the screen's device file doesn't exist, create it with reasonable
        # initial values
        if e.errno == errno.ENOENT:
            fd = io.open(screen_filename(), 'w+b', buffering=0)
            fd.write(b'\x00\x00' * 64)
            fd.write(b''.join(chr(i).encode('ascii') for i in GAMMA_DEFAULT))
        else:
            raise
    return fd


class ScreenClient(object):
    def __init__(self):
        self._fd = init_screen()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_READ)
        self._screen = np.frombuffer(self._map, dtype=np.uint16, count=64).reshape((8, 8))
        self._gamma = np.frombuffer(self._map, dtype=np.uint8, count=32, offset=128)
        self._touch_stop = Event()
        self._touch_thread = Thread(target=self._touch_run)
        self._touch_thread.daemon = True
        self._touch_thread.start()

    def close(self):
        if self._fd:
            self._touch_stop.set()
            self._touch_thread.join()
            self._touch_thread = None
            self._map.close()
            self._map = None
            self._fd.close()
            self._fd = None

    def _touch_run(self):
        while not self._touch_stop.wait(1):
            # "touch" the screen's frame-buffer once a second. This ensures
            # that the screen always updates at least once a second and works
            # around the issue that screen updates can be lost due to lack of
            # resolution the file modification timestamps. Unfortunately,
            # futimes(3) is not universally supported, and only available in
            # Python 3.3+ so this gets a bit convoluted...
            try:
                if os.utime in os.supports_fd:
                    os.utime(self._fd.fileno())
                else:
                    raise NotImplementedError
            except (AttributeError, NotImplementedError) as e:
                os.utime(self._fd.name, None)

    @property
    def array(self):
        return self._screen

    @property
    def rgb_array(self):
        a = np.empty((8, 8, 3), dtype=np.uint8)
        # convert the RGB565 pixels to RGB555 (as the real hardware does)
        a[..., 0] = ((self._screen & 0xF800) >> 11).astype(np.uint8)
        a[..., 1] = ((self._screen & 0x07E0) >> 6).astype(np.uint8)
        a[..., 2] = (self._screen & 0x001F).astype(np.uint8)
        # map all values according to the gamma table
        a = np.take(self._gamma, a)
        # convert to RGB888
        a = a << 3 | a >> 2
        return a

    @property
    def timestamp(self):
        return os.fstat(self._fd.fileno()).st_mtime




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
native_str = str
str = type('')

import io
import os
import sys
import glob
import errno
import struct
import select
import inspect
import socket
from functools import wraps
from collections import namedtuple
from threading import Thread, Event
try:
    from queue import Queue, Empty
except:
    from Queue import Queue, Empty # py2.x
from time import sleep


DIRECTION_UP     = 'up'
DIRECTION_DOWN   = 'down'
DIRECTION_LEFT   = 'left'
DIRECTION_RIGHT  = 'right'
DIRECTION_MIDDLE = 'middle'

ACTION_PRESSED  = 'pressed'
ACTION_RELEASED = 'released'
ACTION_HELD     = 'held'


class InputEvent(namedtuple('InputEvent', ('timestamp', 'direction', 'action'))):
    """
    A :func:`~collections.namedtuple` derivative representing a joystick
    event. The following attributes are present:

    .. attribute:: timestamp

        The time at which the event occurred, represented as the number of
        seconds since the UNIX epoch (same output as :func:`~time.time`).

    .. attribute:: direction

        The direction in which the joystick was pushed (or released from), as
        one of the constants :data:`DIRECTION_UP`, :data:`DIRECTION_DOWN`,
        :data:`DIRECTION_LEFT`, :data:`DIRECTION_RIGHT`,
        :data:`DIRECTION_MIDDLE`

    .. attribute:: action

        The action that occurred, as one of the constants
        :data:`ACTION_PRESSED`, :data:`ACTION_RELEASED`, or
        :data:`ACTION_HELD`.
    """


def stick_address():
    """
    Return the socket address used represent the state of the emulated sense
    HAT's joystick. On UNIX we try ``/dev/shm`` then fall back to ``/tmp``
    (UNIX sockets); on Windows we use localhost.
    """
    fname = 'rpi-sense-emu-stick'
    if sys.platform.startswith('win'):
        # use UDP sockets on Windows
        return (socket.AF_INET, socket.SOCK_DGRAM, ('127.0.0.1', 53753))
    else:
        # use UNIX sockets everywhere else
        if os.path.exists('/dev/shm'):
            return (socket.AF_UNIX, socket.SOCK_DGRAM, os.path.join('/dev/shm', fname))
        else:
            return (socket.AF_UNIX, socket.SOCK_DGRAM, os.path.join('/tmp', fname))


def init_stick_client():
    """
    Opens a socket representing the state of the joystick as a series of evdev
    events. A file-like object is returned (readable like the character device
    representing the real joystick).

    A background thread is spawned to take care of connecting to the stick
    server (and to automatically handle re-connections in the case of
    termination). The thread is marked as a daemon thread so it won't prevent
    script shutdown.
    """
    family, sock_type, addr = stick_address()
    client = socket.socket(family, sock_type)
    if family == socket.AF_INET:
        client.bind(('127.0.0.1', 0))
    elif family == socket.AF_UNIX:
        fname = 'rpi-sense-emu-client-%d' % os.getpid()
        addr_path = os.path.dirname(addr)
        try:
            os.unlink(os.path.join(addr_path, fname))
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        client.bind(os.path.join(addr_path, fname))
    if not client.getsockname():
        raise RuntimeError('Failed to create client socket for stick emulation')
    # Start up a background thread which persistently attempts to connect to
    # the server and pings it with "hello" to notify it that we want stick
    # events. This must be persistent in case the emulating client is stopped
    # and restarted
    def ping_server():
        while True:
            try:
                client.connect(addr)
                client.send(b'hello')
            except socket.error as e:
                if e.errno not in (errno.ENOENT, errno.ENOTCONN, errno.ECONNREFUSED):
                    raise
            sleep(1)
    thread = Thread(target=ping_server)
    thread.daemon = True
    thread.start()
    # Construct file object on top of the socket which we'll return as the
    # result
    return client.makefile('rb', 0)


class SenseStick(object):
    """
    Represents the joystick on the Sense HAT.
    """
    SENSE_HAT_EVDEV_NAME = 'Raspberry Pi Sense HAT Joystick'
    EVENT_FORMAT = native_str('llHHI')
    EVENT_SIZE = struct.calcsize(EVENT_FORMAT)

    EV_KEY = 0x01

    STATE_RELEASE = 0
    STATE_PRESS = 1
    STATE_HOLD = 2

    KEY_UP = 103
    KEY_LEFT = 105
    KEY_RIGHT = 106
    KEY_DOWN = 108
    KEY_ENTER = 28

    def __init__(self):
        self._stick_file = self._stick_device()
        self._callbacks = {}
        self._callback_thread = None
        self._callback_event = Event()

    def close(self):
        if self._stick_file:
            self._callbacks.clear()
            self._start_stop_thread()
            self._stick_file.close()
            self._stick_file = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    def _stick_device(self):
        """
        Discovers the filename of the evdev device that represents the Sense
        HAT's joystick.
        """
        return init_stick_client()

    def _read(self):
        """
        Reads a single event from the joystick, blocking until one is
        available. Returns ``None`` if a non-key event was read, or an
        :class:`InputEvent` tuple describing the event otherwise.
        """
        event = self._stick_file.read(self.EVENT_SIZE)
        (tv_sec, tv_usec, type, code, value) = struct.unpack(self.EVENT_FORMAT, event)
        if type == self.EV_KEY:
            return InputEvent(
                timestamp=tv_sec + (tv_usec / 1000000),
                direction={
                    self.KEY_UP:    DIRECTION_UP,
                    self.KEY_DOWN:  DIRECTION_DOWN,
                    self.KEY_LEFT:  DIRECTION_LEFT,
                    self.KEY_RIGHT: DIRECTION_RIGHT,
                    self.KEY_ENTER: DIRECTION_MIDDLE,
                    }[code],
                action={
                    self.STATE_PRESS:   ACTION_PRESSED,
                    self.STATE_RELEASE: ACTION_RELEASED,
                    self.STATE_HOLD:    ACTION_HELD,
                    }[value])
        else:
            return None

    def _wait(self, timeout=None):
        """
        Waits *timeout* seconds until an event is available from the
        joystick. Returns ``True`` if an event became available, and ``False``
        if the timeout expired.
        """
        r, w, x = select.select([self._stick_file], [], [], timeout)
        return bool(r)

    def _wrap_callback(self, fn):
        # Shamelessley nicked (with some variation) from GPIO Zero :)
        @wraps(fn)
        def wrapper(event):
            return fn()

        if fn is None:
            return None
        elif not callable(fn):
            raise ValueError('value must be None or a callable')
        elif inspect.isbuiltin(fn):
            # We can't introspect the prototype of builtins. In this case we
            # assume that the builtin has no (mandatory) parameters; this is
            # the most reasonable assumption on the basis that pre-existing
            # builtins have no knowledge of InputEvent, and the sole parameter
            # we would pass is an InputEvent
            return wrapper
        else:
            # Try binding ourselves to the argspec of the provided callable.
            # If this works, assume the function is capable of accepting no
            # parameters and that we have to wrap it to ignore the event
            # parameter
            try:
                inspect.getcallargs(fn)
                return wrapper
            except TypeError:
                try:
                    # If the above fails, try binding with a single tuple
                    # parameter. If this works, return the callback as is
                    inspect.getcallargs(fn, ())
                    return fn
                except TypeError:
                    raise ValueError(
                        'value must be a callable which accepts up to one '
                        'mandatory parameter')

    def _start_stop_thread(self):
        if self._callbacks and not self._callback_thread:
            self._callback_event.clear()
            self._callback_thread = Thread(target=self._callback_run)
            self._callback_thread.daemon = True
            self._callback_thread.start()
        elif not self._callbacks and self._callback_thread:
            self._callback_event.set()
            self._callback_thread.join()
            self._callback_thread = None

    def _callback_run(self):
        while not self._callback_event.wait(0):
            event = self._read()
            if event:
                callback = self._callbacks.get(event.direction)
                if callback:
                    callback(event)
                callback = self._callbacks.get('*')
                if callback:
                    callback(event)

    def wait_for_event(self, emptybuffer=False):
        """
        Waits until a joystick event becomes available.  Returns the event, as
        an :class:`InputEvent` tuple.

        If *emptybuffer* is ``True`` (it defaults to ``False``), any pending
        events will be thrown away first. This is most useful if you are only
        interested in "pressed" events.
        """
        if emptybuffer:
            while self._wait(0):
                self._read()
        while self._wait():
            event = self._read()
            if event:
                return event

    def get_events(self):
        """
        Returns a list of all joystick events that have occurred since the last
        call to :meth:`get_events`. The list contains events in the order that
        they occurred. If no events have occurred in the intervening time, the
        result is an empty list.
        """
        result = []
        while self._wait(0):
            event = self._read()
            if event:
                result.append(event)
        return result

    @property
    def direction_up(self):
        """
        The function to be called when the joystick is pushed up. The function
        can either take a parameter which will be the :class:`InputEvent` tuple
        that has occurred, or the function can take no parameters at all.

        Assign ``None`` to prevent this event from being fired.
        """
        return self._callbacks.get(DIRECTION_UP)

    @direction_up.setter
    def direction_up(self, value):
        self._callbacks[DIRECTION_UP] = self._wrap_callback(value)
        self._start_stop_thread()

    @property
    def direction_down(self):
        """
        The function to be called when the joystick is pushed down. The
        function can either take a parameter which will be the
        :class:`InputEvent` tuple that has occurred, or the function can take
        no parameters at all.

        Assign ``None`` to prevent this event from being fired.
        """
        return self._callbacks.get(DIRECTION_DOWN)

    @direction_down.setter
    def direction_down(self, value):
        self._callbacks[DIRECTION_DOWN] = self._wrap_callback(value)
        self._start_stop_thread()

    @property
    def direction_left(self):
        """
        The function to be called when the joystick is pushed left. The
        function can either take a parameter which will be the
        :class:`InputEvent` tuple that has occurred, or the function can take
        no parameters at all.

        Assign ``None`` to prevent this event from being fired.
        """
        return self._callbacks.get(DIRECTION_LEFT)

    @direction_left.setter
    def direction_left(self, value):
        self._callbacks[DIRECTION_LEFT] = self._wrap_callback(value)
        self._start_stop_thread()

    @property
    def direction_right(self):
        """
        The function to be called when the joystick is pushed right. The
        function can either take a parameter which will be the
        :class:`InputEvent` tuple that has occurred, or the function can take
        no parameters at all.

        Assign ``None`` to prevent this event from being fired.
        """
        return self._callbacks.get(DIRECTION_RIGHT)

    @direction_right.setter
    def direction_right(self, value):
        self._callbacks[DIRECTION_RIGHT] = self._wrap_callback(value)
        self._start_stop_thread()

    @property
    def direction_middle(self):
        """
        The function to be called when the joystick middle click is pressed.
        The function can either take a parameter which will be the
        :class:`InputEvent` tuple that has occurred, or the function can take
        no parameters at all.

        Assign ``None`` to prevent this event from being fired.
        """
        return self._callbacks.get(DIRECTION_MIDDLE)

    @direction_middle.setter
    def direction_middle(self, value):
        self._callbacks[DIRECTION_MIDDLE] = self._wrap_callback(value)
        self._start_stop_thread()

    @property
    def direction_any(self):
        """
        The function to be called when the joystick is used. The function can
        either take a parameter which will be the :class:`InputEvent` tuple
        that has occurred, or the function can take no parameters at all.

        This event will always be called *after* events associated with a
        specific action. Assign ``None`` to prevent this event from being
        fired.
        """
        return self._callbacks.get('*')

    @direction_any.setter
    def direction_any(self, value):
        self._callbacks['*'] = self._wrap_callback(value)
        self._start_stop_thread()


class StickServer(object):
    def __init__(self):
        family, sock_type, addr = stick_address()
        server = socket.socket(family, sock_type)
        if family == socket.AF_UNIX:
            try:
                # Kill any pre-existing socket
                os.unlink(addr)
            except OSError:
                pass
        server.bind(addr)
        self._stop = Event()
        self._queue = Queue()
        self._thread = Thread(target=self._serve, args=(server,))
        self._thread.daemon = True
        self._thread.start()

    def _serve(self, server):
        try:
            clients = set()
            while not self._stop.wait(0):
                # Pick up any new clients waiting to receive events
                while select.select([server], [], [], 0)[0]:
                    data, addr = server.recvfrom(64)
                    if data == b'hello':
                        clients.add(addr)
                try:
                    # Grab any data waiting to be sent to clients; we put the
                    # only pause for the thread here to ensure timely response
                    # to events being placed in the queue
                    buf = self._queue.get(timeout=0.1)
                except Empty:
                    pass
                else:
                    # Send the event to all connected clients (pruning any that
                    # fail)
                    for client in list(clients):
                        try:
                            server.sendto(buf, client)
                        except socket.error as e:
                            if e.errno in (errno.ENOENT, errno.ECONNREFUSED):
                                clients.remove(client)
        finally:
            family = server.family
            addr = server.getsockname()
            server.close()
            if family == socket.AF_UNIX:
                # Only works because socket name is guaranteed to be absolute
                os.unlink(addr)

    def close(self):
        if self._thread:
            self._stop.set()
            self._thread.join()

    def send(self, buf):
        self._queue.put(buf)




from sense_emu import *
import numpy as np

# Draw the foreground (fg) into a numpy array
Rd = (255, 0, 0)
Gn = (0, 255, 0)
Bl = (0, 0, 255)
Gy = (128, 128, 128)
__ = (0, 0, 0)
fg = np.array([
    [Rd, Rd, Rd, __, Gn, Gn, __, __],
    [__, Rd, __, __, Gn, __, Gn, __],
    [__, Rd, __, __, Gn, Gn, __, __],
    [__, Rd, __, __, Gn, __, __, __],
    [Bl, __, Bl, __, __, Gy, __, __],
    [Bl, Bl, Bl, __, Gy, __, Gy, __],
    [Bl, __, Bl, __, Gy, __, Gy, __],
    [Bl, __, Bl, __, __, Gy, Gy, __],
    ], dtype=np.uint8)
# Mask is a boolean array of which pixels are transparent
mask = np.all(fg == __, axis=2)

def display(hat, selection):
    # Draw the background (bg) selection box into another numpy array
    left, top, right, bottom = {
        'T': (0, 0, 4, 4),
        'P': (4, 0, 8, 4),
        'Q': (4, 4, 8, 8),
        'H': (0, 4, 4, 8),
        }[selection]
    bg = np.zeros((8, 8, 3), dtype=np.uint8)
    bg[top:bottom, left:right, :] = (255, 255, 255)
    # Construct final pixels from bg array with non-transparent elements of
    # the menu array
    hat.set_pixels([
        bg_pix if mask_pix else fg_pix
        for (bg_pix, mask_pix, fg_pix) in zip(
            (p for row in bg for p in row),
            (p for row in mask for p in row),
            (p for row in fg for p in row),
            )
        ])

def execute(hat, selection):
    if selection == 'T':
        hat.show_message('Temperature: %.1fC' % hat.temp, 0.05, Rd)
    elif selection == 'P':
        hat.show_message('Pressure: %.1fmbar' % hat.pressure, 0.05, Gn)
    elif selection == 'H':
        hat.show_message('Humidity: %.1f%%' % hat.humidity, 0.05, Bl)
    else:
        return True
    return False

def move(selection, direction):
    return {
        ('T', DIRECTION_RIGHT): 'P',
        ('T', DIRECTION_DOWN):  'H',
        ('P', DIRECTION_LEFT):  'T',
        ('P', DIRECTION_DOWN):  'Q',
        ('Q', DIRECTION_UP):    'P',
        ('Q', DIRECTION_LEFT):  'H',
        ('H', DIRECTION_RIGHT): 'Q',
        ('H', DIRECTION_UP):    'T',
        }.get((selection, direction), selection)

hat = SenseHat()
selection = 'T'
while True:
    display(hat, selection)
    event = hat.stick.wait_for_event()
    if event.action == ACTION_PRESSED:
        if event.direction == DIRECTION_MIDDLE:
            if execute(hat, selection):
                break
        else:
            selection = move(selection, event.direction)
hat.clear()




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

"""
Makes Python 2's ConfigParser more like configparser.

This is far from a complete translation, and it's strictly read-only, but it's
enough for my purposes...
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

try:
    import ConfigParser as _ConfigParser
    import collections

    BasicInterpolation = object()

    ExtendedInterpolation = object()

    def ConfigParser(
            defaults=None, dict_type=collections.OrderedDict,
            allow_no_value=False, delimiters=('=', ':'),
            comment_prefixes=('#', ';'), inline_comment_prefixes=None, strict=True,
            empty_lines_in_values=True, default_section='DEFAULT',
            interpolation=BasicInterpolation):
        if not strict:
            raise NotImplementedError
        if not empty_lines_in_values:
            raise NotImplementedError
        if default_section != 'DEFAULT':
            raise NotImplementedError
        if delimiters != ('=', ':'):
            raise NotImplementedError
        if comment_prefixes != ('#', ';'):
            raise NotImplementedError
        if interpolation is None:
            result = _ConfigParser.RawConfigParser(defaults, dict_type, allow_no_value)
        elif interpolation is BasicInterpolation:
            result = _ConfigParser.SafeConfigParser(defaults, dict_type, allow_no_value)
        else:
            raise NotImplementedError
        return result
except ImportError:
    from configparser import (
        ConfigParser,
        BasicInterpolation,
        ExtendedInterpolation,
        )




from sense_emu import SenseHat

sense = SenseHat()

green = (0, 255, 0)
white = (255, 255, 255)

while True:
    humidity = sense.humidity
    humidity_value = 64 * humidity / 100
    pixels = [green if i < humidity_value else white for i in range(64)]
    sense.set_pixels(pixels)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')


import io
import os
import sys
import atexit
import struct
import math
import errno
import subprocess
import webbrowser
import datetime as dt
from time import time, sleep
from threading import Thread, Lock, Event

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GdkPixbuf, Gio, GLib, GObject
import numpy as np
import pkg_resources

from . import __project__, __version__, __author__, __author_email__, __url__
from .i18n import init_i18n, _
from .screen import ScreenClient
from .imu import IMUServer
from .pressure import PressureServer
from .humidity import HumidityServer
from .stick import StickServer, SenseStick
from .lock import EmulatorLock
from .common import HEADER_REC, DATA_REC, DataRecord, slow_pi


def main():
    init_i18n()
    # threads_init isn't required since PyGObject 3.10.2, but just in case
    # we're on something ancient...
    GObject.threads_init()
    app = EmuApplication()
    app.run(sys.argv)


def load_image(filename, format='png'):
    loader = GdkPixbuf.PixbufLoader.new_with_type(format)
    loader.write(pkg_resources.resource_string(__name__, filename))
    loader.close()
    return loader.get_pixbuf()


class EmuApplication(Gtk.Application):
    def __init__(self, *args, **kwargs):
        super(EmuApplication, self).__init__(
                *args, application_id='org.raspberrypi.sense_emu_gui',
                flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
                **kwargs)
        GLib.set_application_name(_('Sense HAT Emulator'))
        self.window = None

    def do_startup(self):
        # super-call needs to be in this form?!
        Gtk.Application.do_startup(self)

        # Get the emulator lock and terminate if something already has it
        self.lock = EmulatorLock('sense_emu_gui')
        try:
            self.lock.acquire()
        except:
            dialog = Gtk.MessageDialog(
                message_type=Gtk.MessageType.ERROR,
                title=_('Error'),
                text=_(
                    'Another process is currently acting as the Sense HAT '
                    'emulator'),
                buttons=Gtk.ButtonsType.CLOSE)
            try:
                dialog.run()
            finally:
                dialog.destroy()
                self.quit()
                return

        def make_action(action_id, handler, param_type=None):
            action = Gio.SimpleAction.new(action_id, param_type)
            action.connect('activate', handler)
            self.add_action(action)
        make_action('example', self.on_example, GLib.VariantType.new('s'))
        make_action('play',    self.on_play)
        make_action('prefs',   self.on_prefs)
        make_action('help',    self.on_help)
        make_action('about',   self.on_about)
        make_action('quit',    self.on_quit)

        builder = Gtk.Builder(translation_domain=__project__)
        builder.add_from_string(
            pkg_resources.resource_string(__name__, 'menu.ui').decode('utf-8'))
        self.props.menubar = builder.get_object('app-menu')

        # Construct the open examples sub-menu
        for directory, label in [
                # I18N: Easy examples
                ('basic',        _('Simple')),
                # I18N: Intermediate skill examples
                ('intermediate', _('Intermediate')),
                # I18N: Difficult examples
                ('advanced',     _('Advanced')),
                ]:
            examples = Gio.Menu.new()
            # NOTE: The use of literal "/" below is correct; resource paths
            # are not file-system paths and always use "/"
            for example in sorted(
                    pkg_resources.resource_listdir(__name__, 'examples/%s' % directory)):
                if example.endswith('.py'):
                    examples.append(
                        example.replace('_', '__'),
                        Gio.Action.print_detailed_name(
                            'app.example',
                            GLib.Variant.new_string('%s/%s' % (directory, example))
                            )
                        )
            builder.get_object('example-submenu').append_submenu(label, examples)

        # Construct the settings database and tweak initial value of
        # simulate-imu and simulate-env if we're running on a slow Pi, and the
        # user hasn't explicitly set a value yet
        if pkg_resources.resource_exists(__name__, 'gschemas.compiled'):
            source = Gio.SettingsSchemaSource.new_from_directory(
                os.path.dirname(pkg_resources.resource_filename(__name__, 'gschemas.compiled')),
                Gio.SettingsSchemaSource.get_default(), True)
        else:
            source = Gio.SettingsSchemaSource.get_default()
        schema = Gio.SettingsSchemaSource.lookup(
            source, self.props.application_id, False)
        assert schema is not None
        self.settings = Gio.Settings.new_full(schema, None, None)
        if self.settings.get_user_value('simulate-imu') is None:
            enable_simulators = not slow_pi()
            self.settings.set_boolean('simulate-imu', enable_simulators)
            self.settings.set_boolean('simulate-env', enable_simulators)

        # Construct the emulator servers
        self.imu = IMUServer(simulate_world=self.settings.get_boolean('simulate-imu'))
        self.pressure = PressureServer(simulate_noise=self.settings.get_boolean('simulate-env'))
        self.humidity = HumidityServer(simulate_noise=self.settings.get_boolean('simulate-env'))
        self.screen = ScreenClient()
        self.stick = StickServer()

        # Connect the settings to the components
        self.settings.connect('changed', self.settings_changed)

    def settings_changed(self, settings, key):
        if key == 'simulate-env':
            self.pressure.simulate_noise = settings.get_boolean(key)
            self.humidity.simulate_noise = settings.get_boolean(key)
        elif key == 'simulate-imu':
            self.imu.simulate_world = settings.get_boolean(key)
        elif key == 'orientation-scale':
            # Force the orientation sliders to redraw
            self.window.ui.yaw_scale.queue_draw()
            self.window.ui.pitch_scale.queue_draw()
            self.window.ui.roll_scale.queue_draw()
        elif key == 'screen-fps':
            self.window.screen_update_delay = 1 / settings.get_int(key)
        else:
            assert False

    def do_shutdown(self):
        if self.lock.mine:
            self.lock.release()
            if self.window:
                self.window.destroy()
                self.window = None
            self.stick.close()
            self.screen.close()
            self.humidity.close()
            self.pressure.close()
            self.imu.close()
        Gtk.Application.do_shutdown(self)

    def do_activate(self):
        if not self.window and self.lock.mine:
            self.window = MainWindow(application=self)
            # Force a read of settings specific to the main window
            self.settings_changed(self.settings, 'screen-fps')
            self.settings_changed(self.settings, 'orientation-scale')
        if self.window:
            self.window.present()

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        # do stuff with switches
        self.activate()
        return 0

    def on_help(self, action, param):
        local_help = '/usr/share/doc/python-sense-emu-doc/html/index.html'
        remote_help = 'https://sense-emu.readthedocs.io/'
        if os.path.exists(local_help):
            webbrowser.open('file://' + local_help)
        else:
            webbrowser.open(remote_help)

    def on_about(self, action, param):
        logo = load_image('sense_emu_gui.svg', format='svg')
        about_dialog = Gtk.AboutDialog(
            transient_for=self.window,
            authors=['%s <%s>' % (__author__, __author_email__)],
            license_type=Gtk.License.GPL_2_0, logo=logo,
            version=__version__, website=__url__)
        about_dialog.run()
        about_dialog.destroy()

    def on_example(self, action, param):
        # NOTE: The use of a bare "/" below is correct: resource paths are
        # *not* file-system paths and always use "/" path separators
        filename = param.unpack()
        source = pkg_resources.resource_stream(
            __name__, '/'.join(('examples', filename)))
        # Write to a filename in the user's home-dir with the timestamp
        # appended to ensure uniqueness (ish)
        filename = os.path.splitext(os.path.basename(filename))[0]
        filename = '{filename}-{timestamp:%Y-%m-%d-%H-%M-%S}.py'.format(
            filename=filename, timestamp=dt.datetime.now())
        filename = os.path.join(os.path.expanduser('~'), filename)
        target = io.open(filename, 'w', encoding='utf-8')
        # Write a note at the top of the file to explain things
        target.write("""\
# This file has been written to your home directory for convenience. It is
# saved as "{filename}"

""".format(filename=filename))
        target.write(source.read().decode('utf-8'))
        # Spawn IDLE; if this seems like a crazy way to spawn IDLE, you're
        # right but it's also cross-platform and cross-version compatible
        # (works on Py 2.x on Windows and UNIX, and Py 3.x on Windows and UNIX;
        # almost any other variant fails for some combination)
        subprocess.Popen([
            sys.executable,
            '-c', 'from idlelib.PyShell import main; main()',
            filename])

    def on_play(self, action, param):
        open_dialog = Gtk.FileChooserDialog(
            transient_for=self.window,
            title=_('Select the recording to play'),
            action=Gtk.FileChooserAction.OPEN)
        open_dialog.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)
        open_dialog.add_button(Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT)
        try:
            response = open_dialog.run()
            open_dialog.hide()
            if response == Gtk.ResponseType.ACCEPT:
                self.window.play(open_dialog.get_filename())
        finally:
            open_dialog.destroy()

    def on_prefs(self, action, param):
        prefs_dialog = PrefsDialog(
            transient_for=self.window,
            title=_('Preferences'),
            settings=self.settings)
        try:
            prefs_dialog.run()
        finally:
            prefs_dialog.destroy()

    def on_quit(self, action, param):
        self.quit()


class BuilderUi(object):
    def __init__(self, owner, filename):
        # Load the GUI definitions (see __getattr__ for how we tie the loaded
        # objects into instance variables) and connect all handlers to methods
        # on this object
        self._builder = Gtk.Builder(translation_domain=__project__)
        self._builder.add_from_string(
            pkg_resources.resource_string(__name__, filename).decode('utf-8'))
        self._builder.connect_signals(owner)

    def __getattr__(self, name):
        result = self._builder.get_object(name)
        if result is None:
            raise AttributeError(_('No such attribute %r') % name)
        setattr(self, name, result)
        return result


class MainWindow(Gtk.ApplicationWindow):
    def __init__(self, *args, **kwargs):
        super(MainWindow, self).__init__(*args, **kwargs)

        # Build the UI; this is a bit round-about because of Gtk's weird UI
        # handling. One can't just use a UI file to setup an existing Window
        # instance (as in Qt); instead one must use a separate handler object
        # (in which case overriding do_destroy is impossible) or construct a
        # whole new Window, remove its components, add them to ourselves and
        # then ditch the Window.
        self._ui = BuilderUi(self, 'main_window.ui')
        self.ui.window.remove(self.ui.root_grid)
        self.add(self.ui.root_grid)
        self.ui.window.destroy()

        # Set the window icon
        icon = load_image('sense_emu_gui.png')
        self.props.icon = icon

        # Set up the objects for the playback thread
        self._play_update_lock = Lock()
        self._play_update_id = 0
        self._play_event = Event()
        self._play_thread = None
        self._play_restore = (True, True, True)

        # Set initial positions on sliders (and add some marks)
        self.ui.pitch_scale.add_mark(0, Gtk.PositionType.BOTTOM, None)
        self.ui.roll_scale.add_mark(0, Gtk.PositionType.BOTTOM, None)
        self.ui.yaw_scale.add_mark(0, Gtk.PositionType.BOTTOM, None)
        self.ui.roll.props.value = self.props.application.imu.orientation[0]
        self.ui.pitch.props.value = self.props.application.imu.orientation[1]
        self.ui.yaw.props.value = self.props.application.imu.orientation[2]
        self.ui.humidity.props.value = self.props.application.humidity.humidity
        self.ui.pressure.props.value = self.props.application.pressure.pressure
        self.ui.temperature.props.value = self.props.application.humidity.temperature

        # Load graphics assets
        self.sense_image = load_image('sense_emu.png')
        self.orient_image = load_image('orientation.png')
        self.pixel_grid = load_image('pixel_grid.png')

        # Set up attributes for the joystick buttons
        self._stick_held_lock = Lock()
        self._stick_held_id = 0
        self.ui.left_button.direction = SenseStick.KEY_LEFT
        self.ui.right_button.direction = SenseStick.KEY_RIGHT
        self.ui.up_button.direction = SenseStick.KEY_UP
        self.ui.down_button.direction = SenseStick.KEY_DOWN
        self.ui.enter_button.direction = SenseStick.KEY_ENTER
        self._stick_map = {
            Gdk.KEY_Return: self.ui.enter_button,
            Gdk.KEY_Left:   self.ui.left_button,
            Gdk.KEY_Right:  self.ui.right_button,
            Gdk.KEY_Up:     self.ui.up_button,
            Gdk.KEY_Down:   self.ui.down_button,
            }

        # Set up attributes for the screen rotation controls
        self.ui.screen_rotate_clockwise.angle = -90
        self.ui.screen_rotate_anticlockwise.angle = 90
        self._stick_rotations = {
            SenseStick.KEY_LEFT:  SenseStick.KEY_UP,
            SenseStick.KEY_UP:    SenseStick.KEY_RIGHT,
            SenseStick.KEY_RIGHT: SenseStick.KEY_DOWN,
            SenseStick.KEY_DOWN:  SenseStick.KEY_LEFT,
            SenseStick.KEY_ENTER: SenseStick.KEY_ENTER,
            }

        # Set up a thread to constantly refresh the pixels from the screen
        # client object
        self.screen_update_delay = 0.04
        self._screen_rotation = 0
        self._screen_orientation = False
        self._screen_pending = False
        self._screen_timestamp = 0.0
        self._screen_event = Event()
        self._screen_thread = Thread(target=self._screen_run)
        self._screen_thread.daemon = True
        self._screen_thread.start()

    @property
    def ui(self):
        return self._ui

    def do_destroy(self):
        try:
            self._play_stop()
            self._screen_event.set()
            self._screen_thread.join()
        except AttributeError:
            # do_destroy gets called multiple times, and subsequent times lacks
            # the Python-added instance attributes
            pass
        Gtk.ApplicationWindow.do_destroy(self)

    def format_pressure(self, scale, value):
        return '%.1fmbar' % value

    def pressure_changed(self, adjustment):
        if not self._play_thread:
            self.props.application.pressure.set_values(
                self.ui.pressure.props.value,
                self.ui.temperature.props.value,
                )

    def format_humidity(self, scale, value):
        return '%.1f%%' % value

    def humidity_changed(self, adjustment):
        if not self._play_thread:
            self.props.application.humidity.set_values(
                self.ui.humidity.props.value,
                self.ui.temperature.props.value,
                )

    def format_temperature(self, scale, value):
        return '%.1f°C' % value

    def temperature_changed(self, adjustment):
        if not self._play_thread:
            self.pressure_changed(adjustment)
            self.humidity_changed(adjustment)

    def format_orientation(self, scale, value):
        mode = self.props.application.settings.get_string('orientation-scale')
        return '%.1f°' % (
            value       if mode == 'balance' else
            value + 180 if mode == 'circle' else
            value % 360 if mode == 'modulo' else
            999 # should never happen
            )

    def orientation_changed(self, adjustment):
        if not self._play_thread:
            self.props.application.imu.set_orientation((
                self.ui.roll.props.value,
                self.ui.pitch.props.value,
                self.ui.yaw.props.value,
                ))

    def stick_key_pressed(self, button, event):
        try:
            button = self._stick_map[event.keyval]
        except KeyError:
            return False
        else:
            self.stick_pressed(button, event)
            return True

    def stick_key_released(self, button, event):
        try:
            button = self._stick_map[event.keyval]
        except KeyError:
            return False
        else:
            self.stick_released(button, event)
            return True

    def stick_pressed(self, button, event):
        # When a button is double-clicked, GTK fires two pressed events for the
        # second click with no intervening released event (so there's one
        # pressed event for the first click, followed by a released event, then
        # two pressed events for the second click followed by a single released
        # event). This isn't documented, so it could be a bug, but it seems
        # more like a deliberate behaviour. Anyway, we work around the
        # redundant press by detecting it with the non-zero stick_held_id and
        # ignoring the redundant event
        button.grab_focus()
        with self._stick_held_lock:
            if self._stick_held_id:
                return True
            self._stick_held_id = GLib.timeout_add(250, self.stick_held_first, button)
        self._stick_send(button.direction, SenseStick.STATE_PRESS)
        button.set_active(True)
        return True

    def stick_released(self, button, event):
        with self._stick_held_lock:
            if self._stick_held_id:
                GLib.source_remove(self._stick_held_id)
                self._stick_held_id = 0
        self._stick_send(button.direction, SenseStick.STATE_RELEASE)
        button.set_active(False)
        return True

    def stick_held_first(self, button):
        with self._stick_held_lock:
            self._stick_held_id = GLib.timeout_add(50, self.stick_held, button)
        self._stick_send(button.direction, SenseStick.STATE_HOLD)
        return False

    def stick_held(self, button):
        self._stick_send(button.direction, SenseStick.STATE_HOLD)
        return True

    def _stick_send(self, direction, action):
        tv_usec, tv_sec = math.modf(time())
        tv_usec *= 1000000
        r = self._screen_rotation // 90
        while r:
            direction = self._stick_rotations[direction]
            r -= 1
        event_rec = struct.pack(SenseStick.EVENT_FORMAT,
            int(tv_sec), int(tv_usec), SenseStick.EV_KEY, direction, action)
        self.props.application.stick.send(event_rec)

    def rotate_screen(self, button):
        self._screen_rotation = (self._screen_rotation + button.angle) % 360
        self.ui.screen_rotate_label.props.label = '%d°' % self._screen_rotation
        self._screen_timestamp = 0 # force a screen update

    def toggle_orientation(self, button):
        self._screen_orientation = not self._screen_orientation
        self._screen_timestamp = 0 # force a screen update

    def _screen_run(self):
        # This method runs in the background _screen_thread
        while True:
            # Only update the screen if there's no idle callback pending from
            # a prior update
            if not self._screen_pending:
                # Only update if the screen's modification timestamp indicates
                # that the data has changed since last time
                ts = self.props.application.screen.timestamp
                if ts > self._screen_timestamp:
                    img = self.sense_image.copy()
                    pixels = GdkPixbuf.Pixbuf.new_from_bytes(
                        GLib.Bytes.new(self.props.application.screen.rgb_array.tostring()),
                        colorspace=GdkPixbuf.Colorspace.RGB, has_alpha=False,
                        bits_per_sample=8, width=8, height=8, rowstride=8 * 3)
                    pixels.composite(img, 31, 38, 128, 128, 31, 38, 16, 16,
                        GdkPixbuf.InterpType.NEAREST, 255)
                    self.pixel_grid.composite(img, 31, 38, 128, 128, 31, 38, 1, 1,
                        GdkPixbuf.InterpType.NEAREST, 255)
                    if self._screen_orientation:
                        self.orient_image.composite(img, 0, 0, 265, 204, 0, 0, 1, 1,
                            GdkPixbuf.InterpType.NEAREST, 215)
                    img = img.rotate_simple(self._screen_rotation)
                    self._screen_pending = True
                    self._screen_timestamp = ts
                    # GTK updates must be done by the main thread; schedule
                    # the image to be redrawn when the app is idle. It would
                    # be better to use custom signals for this ... but then
                    # pixman region copy bugs start appearing
                    GLib.idle_add(self._update_screen, img)
            # The following wait enforces the maximum update rate (set from
            # the preferences dialog)
            if self._screen_event.wait(self.screen_update_delay):
                break

    def _update_screen(self, pixbuf):
        self.ui.screen_image.set_from_pixbuf(pixbuf)
        self._screen_pending = False
        return False

    def _play_run(self, f):
        err = None
        try:
            # Calculate how many records are in the file; we'll use this later
            # when updating the progress bar
            rec_total = (f.seek(0, io.SEEK_END) - HEADER_REC.size) // DATA_REC.size
            f.seek(0)
            skipped = 0
            for rec, data in enumerate(self._play_source(f)):
                now = time()
                if data.timestamp < now:
                    skipped += 1
                    continue
                else:
                    if self._play_event.wait(data.timestamp - now):
                        break
                self.props.application.pressure.set_values(data.pressure, data.ptemp)
                self.props.application.humidity.set_values(data.humidity, data.htemp)
                self.props.application.imu.set_imu_values(
                    (data.ax, data.ay, data.az),
                    (data.gx, data.gy, data.gz),
                    (data.cx, data.cy, data.cz),
                    (data.ox, data.oy, data.oz),
                    )
                # Again, would be better to use custom signals here but
                # attempting to do so just results in seemingly random
                # segfaults during playback
                with self._play_update_lock:
                    if self._play_update_id == 0:
                        self._play_update_id = GLib.idle_add(self._play_update_controls, rec / rec_total)
        except Exception as e:
            err = e
        finally:
            f.close()
            # Must ensure that controls are only re-enabled *after* all pending
            # control updates have run
            with self._play_update_lock:
                if self._play_update_id:
                    GLib.source_remove(self._play_update_id)
                    self._play_update_id = 0
            # Get the main thread to re-enable the controls at the end of
            # playback
            GLib.idle_add(self._play_controls_finish, err)

    def _play_update_controls(self, fraction):
        with self._play_update_lock:
            self._play_update_id = 0
        self.ui.play_progressbar.props.fraction = fraction
        if not math.isnan(self.props.application.humidity.temperature):
            self.ui.temperature.props.value = self.props.application.humidity.temperature
        if not math.isnan(self.props.application.pressure.pressure):
            self.ui.pressure.props.value = self.props.application.pressure.pressure
        if not math.isnan(self.props.application.humidity.humidity):
            self.ui.humidity.props.value = self.props.application.humidity.humidity
        self.ui.yaw.props.value = math.degrees(self.props.application.imu.orientation[2])
        self.ui.pitch.props.value = math.degrees(self.props.application.imu.orientation[1])
        self.ui.roll.props.value = math.degrees(self.props.application.imu.orientation[0])
        return False

    def play_stop_clicked(self, button):
        self._play_stop()

    def _play_stop(self):
        if self._play_thread:
            self._play_event.set()
            self._play_thread.join()
            self._play_thread = None

    def _play_source(self, f):
        magic, ver, offset = HEADER_REC.unpack(f.read(HEADER_REC.size))
        if magic != b'SENSEHAT':
            raise IOError(_('%s is not a Sense HAT recording') % f.name)
        if ver != 1:
            raise IOError(_('%s has unrecognized file version number') % f.name)
        offset = time() - offset
        while True:
            buf = f.read(DATA_REC.size)
            if not buf:
                break
            elif len(buf) < DATA_REC.size:
                raise IOError(_('Incomplete data record at end of %s') % f.name)
            else:
                data = DataRecord(*DATA_REC.unpack(buf))
                yield data._replace(timestamp=data.timestamp + offset)

    def _play_controls_setup(self, filename):
        # Disable all the associated user controls while playing back
        self.ui.environ_box.props.sensitive = False
        self.ui.gyro_grid.props.sensitive = False
        # Disable simulation threads as we're going to manipulate the
        # values precisely
        self._play_restore = (
            self.props.application.pressure.simulate_noise,
            self.props.application.humidity.simulate_noise,
            self.props.application.imu.simulate_world,
            )
        self.props.application.pressure.simulate_noise = False
        self.props.application.humidity.simulate_noise = False
        self.props.application.imu.simulate_world = False
        # Show the playback bar
        self.ui.play_label.props.label = _('Playing %s') % os.path.basename(filename)
        self.ui.play_progressbar.props.fraction = 0.0
        self.ui.play_box.props.visible = True

    def _play_controls_finish(self, exc):
        # Reverse _play_controls_setup
        self.ui.play_box.props.visible = False
        ( self.props.application.pressure.simulate_noise,
            self.props.application.humidity.simulate_noise,
            self.props.application.imu.simulate_world,
            ) = self._play_restore
        self.ui.environ_box.props.sensitive = True
        self.ui.gyro_grid.props.sensitive = True
        self._play_thread = None
        # If an exception occurred in the background thread, display the
        # error in an appropriate dialog
        if exc:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                message_type=Gtk.MessageType.ERROR,
                title=_('Error'),
                text=_('Error while replaying recording'),
                buttons=Gtk.ButtonsType.CLOSE)
            dialog.format_secondary_text(str(exc))
            dialog.run()
            dialog.destroy()

    def play(self, filename):
        self._play_stop()
        self._play_controls_setup(filename)
        self._play_thread = Thread(target=self._play_run, args=(io.open(filename, 'rb'),))
        self._play_event.clear()
        self._play_thread.start()


class PrefsDialog(Gtk.Dialog):
    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings')
        super(PrefsDialog, self).__init__(*args, **kwargs)

        # See comments in MainWindow...
        self._ui = BuilderUi(self, 'prefs_dialog.ui')
        self.ui.window.remove(self.ui.dialog_vbox)
        self.remove(self.get_content_area())
        self.add(self.ui.dialog_vbox)
        self.ui.window.destroy()

        self.props.resizable = False
        self.ui.close_button.grab_default()
        self.settings.bind(
            'simulate-env', self.ui.env_check, 'active', Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind(
            'simulate-imu', self.ui.imu_check, 'active', Gio.SettingsBindFlags.DEFAULT)
        self.settings.bind(
            'screen-fps', self.ui.screen_fps, 'value', Gio.SettingsBindFlags.DEFAULT)
        self.ui.orientation_balance.value = 'balance'
        self.ui.orientation_circle.value = 'circle'
        self.ui.orientation_modulo.value = 'modulo'
        s = self.settings.get_string('orientation-scale')
        for c in self.ui.orientation_balance.get_group():
            c.props.active = (c.value == s)

    @property
    def ui(self):
        return self._ui

    def close_clicked(self, button):
        self.response(Gtk.ResponseType.ACCEPT)

    def orientation_changed(self, button):
        if button.props.active:
            self.settings.set_string('orientation-scale', button.value)




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')

import io
import errno
from struct import Struct
from collections import namedtuple


# Structures for sense_rec and sense_play
HEADER_REC = Struct(nstr(
    '='  # native order, standard sizing
    '8s' # magic number ("SENSEHAT")
    'b'  # version number (1)
    '7x' # padding
    'd'  # initial timestamp
    ))

DATA_REC = Struct(nstr(
    '='   # native order, standard sizing
    'd'   # timestamp
    'dd'  # pressure+temp readings
    'dd'  # humidity+temp readings
    'ddd' # raw accelerometer readings
    'ddd' # raw gyro readings
    'ddd' # raw compass readings
    'ddd' # calculated pose
    ))

DataRecord = namedtuple('DataRecord', (
    'timestamp',
    'pressure', 'ptemp',
    'humidity', 'htemp',
    'ax', 'ay', 'az',
    'gx', 'gy', 'gz',
    'cx', 'cy', 'cz',
    'ox', 'oy', 'oz',
    ))


def clamp(value, min_value, max_value):
    """
    Return *value* clipped to the range *min_value* to *max_value* inclusive.
    """
    return min(max_value, max(min_value, value))


def slow_pi():
    """
    Returns ``True`` if the local hardware is a Raspberry Pi with a slow
    processor, specifically a BCM2835. This is used to determine defaults for
    the simulation's processing.
    """
    try:
        cpu = ''
        with io.open('/proc/cpuinfo', 'r') as f:
            for line in f:
                if line.startswith('Hardware'):
                    cpu = line.split(':', 1)[1].strip()
                    break
        return cpu in ('BCM2835', 'BCM2708')
    except IOError as e:
        if e.errno == errno.ENOENT:
            return False
        raise




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import sys
import locale
import gettext as _gettext
import atexit

import pkg_resources

from . import __project__

def init_i18n(languages=None):
    # Ensure any resources we extract get cleaned up interpreter shutdown
    atexit.register(pkg_resources.cleanup_resources)
    # Figure out where the language catalogs are; this will extract them
    # if the package is frozen
    localedir = pkg_resources.resource_filename(__name__, 'locale')
    try:
        # Use the user's default locale instead of C
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as e:
        # If locale is not supported, use C which should at least provide
        # consistency. In this case, don't set a gettext domain to prevent
        # translation of strings
        locale.setlocale(locale.LC_ALL, 'C')
    else:
        # Set translation domain for GNU's gettext (needed by GTK's Builder)
        try:
            locale.bindtextdomain(__project__, localedir)
            locale.textdomain(__project__)
        except AttributeError:
            if sys.platform.startswith('win'):
                try:
                    # We're on Windows; try and use intl.dll instead
                    import ctypes
                    libintl = ctypes.cdll.LoadLibrary('intl.dll')
                except OSError:
                    # intl.dll isn't available; give up
                    return
                else:
                    libintl.bindtextdomain(__project__, localedir)
                    libintl.textdomain(__project__)
                    libintl.bind_textdomain_codeset(__project, 'UTF-8')
            else:
                # We're on something else (Mac OS X most likely); no idea what
                # to do here yet
                return
        # Finally, set translation domain for Python's built-in gettext
        _gettext.bindtextdomain(__project__, localedir)
        _gettext.textdomain(__project__)


if sys.version_info.major == 3:
    gettext = _gettext.gettext
    ngettext = _gettext.ngettext
else:
    gettext = lambda message: _gettext.gettext(message).decode('utf-8')
    ngettext = lambda singular, plural, n: _gettext.ngettext(singular, plural, n).decode('utf-8')
_ = gettext




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')

import sys
import os
import io
import mmap
import errno
from struct import Struct
from collections import namedtuple
from random import Random
from time import time
from threading import Thread, Event
from math import isnan

import numpy as np

from .common import clamp


# See LPS25H data-sheet for details of register values
PRESSURE_FACTOR = 4096
TEMP_OFFSET = 37
TEMP_FACTOR = 480
PRESSURE_DATA = Struct(nstr(
    '@'   # native mode
    'B'   # pressure sensor type
    '6p'  # pressure sensor name
    'l'   # P_REF
    'l'   # P_OUT
    'h'   # T_OUT
    'B'   # P_VALID
    'B'   # T_VALID
    ))

PressureData = namedtuple('PressureData',
    ('type', 'name', 'P_REF', 'P_OUT', 'T_OUT', 'P_VALID', 'T_VALID'))


def pressure_filename():
    """
    Return the filename used to represent the state of the emulated sense HAT's
    pressure sensor. On UNIX we try ``/dev/shm`` then fall back to ``/tmp``; on
    Windows we use whatever ``%TEMP%`` contains
    """
    fname = 'rpi-sense-emu-pressure'
    if sys.platform.startswith('win'):
        # just use a temporary file on Windows
        return os.path.join(os.environ['TEMP'], fname)
    else:
        if os.path.exists('/dev/shm'):
            return os.path.join('/dev/shm', fname)
        else:
            return os.path.join('/tmp', fname)


def init_pressure():
    """
    Opens the file representing the state of the pressure sensors. The
    file-like object is returned.

    If the file already exists we simply make sure it is the right size. If
    the file does not already exist, it is created and zeroed.
    """
    try:
        # Attempt to open the IMU's device file and ensure it's the right size
        fd = io.open(pressure_filename(), 'r+b', buffering=0)
        fd.seek(PRESSURE_DATA.size)
        fd.truncate()
    except IOError as e:
        # If the screen's device file doesn't exist, create it with reasonable
        # initial values
        if e.errno == errno.ENOENT:
            fd = io.open(pressure_filename(), 'w+b', buffering=0)
            fd.write(b'\x00' * PRESSURE_DATA.size)
        else:
            raise
    return fd


class PressureServer(object):
    def __init__(self, simulate_noise=True):
        self._random = Random()
        self._fd = init_pressure()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_WRITE)
        data = self._read()
        if data.type != 3:
            self._write(PressureData(3, b'LPS25H', 0, 0, 0, 0, 0))
            self._pressure = 1013.0
            self._temperature = 20.0
        else:
            self._pressure = data.P_OUT / 4096
            self._temperature = data.T_OUT / 480 + 42.5
        self._noise_thread = None
        self._noise_event = Event()
        self._noise_write()
        # The queue lengths are selected to accurately represent the response
        # time of the sensors
        self._pressures = np.full((25,), self._pressure, dtype=np.float)
        self._temperatures = np.full((25,), self._temperature, dtype=np.float)
        self.simulate_noise = simulate_noise

    def close(self):
        if self._fd:
            self.simulate_noise = False
            self._map.close()
            self._fd.close()
            self._fd = None
            self._map = None

    def _perturb(self, value, error):
        """
        Return *value* perturbed by +/- *error* which is derived from a
        gaussian random generator.
        """
        # We use an internal Random() instance here to avoid a threading issue
        # with the gaussian generator (could use locks, but an instance of
        # Random is easier and faster)
        return value + self._random.gauss(0, 0.2) * error

    def _read(self):
        return PressureData(*PRESSURE_DATA.unpack_from(self._map))

    def _write(self, value):
        PRESSURE_DATA.pack_into(self._map, 0, *value)

    @property
    def pressure(self):
        return self._pressure

    @property
    def temperature(self):
        return self._temperature

    def set_values(self, pressure, temperature):
        self._pressure = pressure
        self._temperature = temperature
        if not self._noise_thread:
            self._noise_write()

    @property
    def simulate_noise(self):
        return self._noise_thread is not None

    @simulate_noise.setter
    def simulate_noise(self, value):
        if value and not self._noise_thread:
            self._noise_event.clear()
            self._noise_thread = Thread(target=self._noise_loop)
            self._noise_thread.daemon = True
            self._noise_thread.start()
        elif self._noise_thread and not value:
            self._noise_event.set()
            self._noise_thread.join()
            self._noise_thread = None
            self._noise_write()

    def _noise_loop(self):
        while not self._noise_event.wait(0.04):
            self._noise_write()

    def _noise_write(self):
        if self.simulate_noise:
            self._pressures[1:] = self._pressures[:-1]
            self._pressures[0] = self._perturb(self.pressure, (
                0.2 if 800 <= self.pressure <= 1100 and 20 <= self.temperature <= 60 else
                1.0))
            self._temperatures[1:] = self._temperatures[:-1]
            self._temperatures[0] = self._perturb(self.temperature, (
                2.0 if 0 <= self.temperature <= 65 else
                4.0))
            pressure = self._pressures.mean()
            temperature = self._temperatures.mean()
        else:
            pressure = self.pressure
            temperature = self.temperature
        self._write(self._read()._replace(
            P_VALID=not isnan(pressure),
            T_VALID=not isnan(temperature),
            P_OUT=0 if isnan(pressure) else int(clamp(pressure, 260, 1260) * PRESSURE_FACTOR),
            T_OUT=0 if isnan(temperature) else int((clamp(temperature, -30, 105) - TEMP_OFFSET) * TEMP_FACTOR),
            ))





from colorsys import hsv_to_rgb
from time import sleep
from sense_emu import SenseHat

# Hues represent the spectrum of colors as values between 0 and 1. The range
# is circular so 0 represents red, ~0.2 is yellow, ~0.33 is green, 0.5 is cyan,
# ~0.66 is blue, ~0.84 is purple, and 1.0 is back to red. These are the initial
# hues for each pixel in the display.
hues = [
    0.00, 0.00, 0.06, 0.13, 0.20, 0.27, 0.34, 0.41,
    0.00, 0.06, 0.13, 0.21, 0.28, 0.35, 0.42, 0.49,
    0.07, 0.14, 0.21, 0.28, 0.35, 0.42, 0.50, 0.57,
    0.15, 0.22, 0.29, 0.36, 0.43, 0.50, 0.57, 0.64,
    0.22, 0.29, 0.36, 0.44, 0.51, 0.58, 0.65, 0.72,
    0.30, 0.37, 0.44, 0.51, 0.58, 0.66, 0.73, 0.80,
    0.38, 0.45, 0.52, 0.59, 0.66, 0.73, 0.80, 0.87,
    0.45, 0.52, 0.60, 0.67, 0.74, 0.81, 0.88, 0.95,
    ]

hat = SenseHat()

def scale(v):
    return int(v * 255)

while True:
    # Rotate the hues
    hues = [(h + 0.01) % 1.0 for h in hues]
    # Convert the hues to RGB values
    pixels = [hsv_to_rgb(h, 1.0, 1.0) for h in hues]
    # hsv_to_rgb returns 0..1 floats; convert to ints in the range 0..255
    pixels = [(scale(r), scale(g), scale(b)) for r, g, b in pixels]
    # Update the display
    hat.set_pixels(pixels)
    sleep(0.04)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')

import sys
import os
import io
import mmap
import time
import errno
import subprocess
from random import Random
from struct import Struct
from collections import namedtuple
from threading import Thread, Event

import numpy as np

from .common import clamp


# See LSM9DS1 data-sheet for details of register values
ACCEL_FACTOR = 4081.6327
GYRO_FACTOR = 57.142857
COMPASS_FACTOR = 7142.8571
ORIENT_FACTOR = 5214.1892
IMU_DATA = Struct(nstr(
    '@'   # native mode
    'B'   # IMU sensor type
    '20p' # IMU sensor name
    'Q'   # timestamp
    'hhh' # OUT_X_G, OUT_Y_G, OUT_Z_G
    'hhh' # OUT_X_XL, OUT_Y_XL, OUT_Z_XL
    'hhh' # OUT_X_M, OUT_Y_M, OUT_Z_M
    'hhh' # Orientation X, Y, Z
    ))

IMUData = namedtuple('IMUData', ('type', 'name', 'timestamp', 'accel', 'gyro', 'compass', 'orient'))


def imu_filename():
    """
    Return the filename used to represent the state of the emulated sense HAT's
    IMU sensors. On UNIX we try ``/dev/shm`` then fall back to ``/tmp``; on
    Windows we use whatever ``%TEMP%`` contains
    """
    fname = 'rpi-sense-emu-imu'
    if sys.platform.startswith('win'):
        # just use a temporary file on Windows
        return os.path.join(os.environ['TEMP'], fname)
    else:
        if os.path.exists('/dev/shm'):
            return os.path.join('/dev/shm', fname)
        else:
            return os.path.join('/tmp', fname)


def init_imu():
    """
    Opens the file representing the state of the IMU sensors. The file-like
    object is returned.

    If the file already exists we simply make sure it is the right size. If
    the file does not already exist, it is created and zeroed.
    """
    try:
        # Attempt to open the IMU's device file and ensure it's the right size
        fd = io.open(imu_filename(), 'r+b', buffering=0)
        fd.seek(IMU_DATA.size)
        fd.truncate()
    except IOError as e:
        # If the screen's device file doesn't exist, create it with reasonable
        # initial values
        if e.errno == errno.ENOENT:
            fd = io.open(imu_filename(), 'w+b', buffering=0)
            fd.write(b'\x00' * IMU_DATA.size)
        else:
            raise
    return fd


# Find the best available time-source for the timestamp() function. The best
# source will preferably be monotonic, and high-resolution
try:
    _time = time.monotonic # 3.3+ (only guaranteed in 3.5+)
except AttributeError:
    try:
        _time = time.perf_counter # 3.3+
    except AttributeError:
        _time = time.time # fallback for 2.7
def timestamp():
    """
    Returns a timestamp as an integer number of microseconds after some
    arbitrary basis (only comparisons of consecutive calls are meaningful).
    """
    return int(_time() * 1000000)


# Some handy array definitions
V = lambda x, y, z: np.array((x, y, z))
O = V(0, 0, 0)
X = V(1, 0, 0)
Y = V(0, 1, 0)
Z = V(0, 0, 1)


class IMUServer(object):
    def __init__(self, simulate_world=True):
        self._random = Random()
        self._fd = init_imu()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_WRITE)
        data = self._read()
        self._gravity = Z
        self._north = 0.33 * X
        if data.type != 6:
            self._write(IMUData(6, b'LSM9DS1', timestamp(), O, O, O, O))
            self._accel = O
            self._gyro = O
            self._compass = O
            self._orientation = O
            self._position = O
        else:
            self._accel = data.accel / ACCEL_FACTOR
            self._gyro = data.gyro / GYRO_FACTOR
            self._compass = data.compass / COMPASS_FACTOR
            self._orientation = O # XXX calc orientation from accel and gravity
            self._position = O # XXX calc position from compass and north
        self._world_thread = None
        self._world_event = Event()
        self._world_write(timestamp())
        # These queue lengths were arbitrarily selected to smooth the action of
        # the orientation sliders in the GUI; they bear no particular relation
        # to the hardware
        self._gyros = np.full((10, 3), self._gyro, dtype=np.float)
        self._accels = np.full((10, 3), self._accel, dtype=np.float)
        self._comps = np.full((10, 3), self._compass, dtype=np.float)
        self.simulate_world = simulate_world

    def close(self):
        if self._fd:
            self.simulate_world = False
            self._map.close()
            self._fd.close()
            self._fd = None
            self._map = None

    def _read(self):
        (
            type, name, timestamp,
            ax, ay, az,
            gx, gy, gz,
            cx, cy, cz,
            ox, oy, oz,
            ) = IMU_DATA.unpack_from(self._map)
        return IMUData(
            type, name, timestamp,
            V(ax, ay, az),
            V(gx, gy, gz),
            V(cx, cy, cz),
            V(ox, oy, oz),
            )

    def _write(self, value):
        value = (
            value.type, value.name, value.timestamp,
            value.accel[0], value.accel[1], value.accel[2],
            value.gyro[0], value.gyro[1], value.gyro[2],
            value.compass[0], value.compass[1], value.compass[2],
            value.orient[0], value.orient[1], value.orient[2],
            )
        IMU_DATA.pack_into(self._map, 0, *value)

    def _perturb(self, value, error):
        """
        Return *value* perturbed by +/- *error* which is derived from a
        gaussian random generator.
        """
        # We use an internal Random() instance here to avoid a threading issue
        # with the gaussian generator (could use locks, but an instance of
        # Random is easier and faster)
        return V(
            value[0] + self._random.gauss(0, 0.2) * error,
            value[1] + self._random.gauss(0, 0.2) * error,
            value[2] + self._random.gauss(0, 0.2) * error,
            )

    def set_orientation(self, orientation, position=None):
        assert self.simulate_world
        if position is None:
            position = O
        self._orientation = V(*orientation)
        self._position = V(*position)

    def set_imu_values(self, accel, gyro, compass, orientation, position=None):
        assert not self.simulate_world
        self._accel = V(*accel)
        self._gyro = V(*gyro)
        self._compass = V(*compass)
        self._orientation = V(*orientation)
        if position is None:
            position = O
        self._position = V(*position)
        self._world_write(timestamp())

    @property
    def accel(self):
        return self._accel

    @property
    def gyro(self):
        return self._gyro

    @property
    def compass(self):
        return self._compass

    @property
    def orientation(self):
        return self._orientation

    @property
    def position(self):
        return self._position

    @property
    def simulate_world(self):
        return self._world_thread is not None

    @simulate_world.setter
    def simulate_world(self, value):
        if value and not self._world_thread:
            self._world_event.clear()
            self._world_thread = Thread(target=self._world_loop)
            self._world_thread.daemon = True
            self._world_thread.start()
        elif self._world_thread and not value:
            self._world_event.set()
            self._world_thread.join()
            self._world_thread = None
            self._world_write(timestamp())

    def _world_loop(self):
        old_timestamp = timestamp()
        old_position = self._position
        old_orientation = self._orientation
        while not self._world_event.wait(0.016):
            # Gyro reading is simply the rate of change of the orientation
            new_timestamp = timestamp()
            new_position = self._position
            new_orientation = self._orientation
            time_delta = (new_timestamp - old_timestamp) / 1000000
            self._gyro = (new_orientation - old_orientation) / time_delta
            # Construct a rotation matrix for the orientation; see
            # https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix
            x, y, z = np.deg2rad(new_orientation)
            c1, c2, c3 = np.cos((z, y, x))
            s1, s2, s3 = np.sin((z, y, x))
            R = np.array([
                [c1 * c2, c1 * s2 * s3 - c3 * s1, s1 * s3 + c1 * c3 * s2],
                [c2 * s1, c1 * c3 + s1 * s2 * s3, c3 * s1 * s2 - c1 * s3],
                [-s2,     c2 * s3,                c2 * c3],
                ])
            self._accel = R.T.dot(self._gravity) # transpose for passive rotation
            self._compass = R.T.dot(self._north)
            # XXX Simulate acceleration from position
            self._world_write(new_timestamp)
            old_timestamp = new_timestamp
            old_position = new_position
            old_orientation = new_orientation

    def _world_write(self, timestamp):
        if self.simulate_world:
            self._gyros[1:, :] = self._gyros[:-1, :]
            self._gyros[0, :] = self._perturb(self._gyro, 1.0)
            gyro = self._gyros.mean(axis=0)
            self._accels[1:, :] = self._accels[:-1, :]
            self._accels[0, :] = self._perturb(self._accel, 0.1)
            accel = self._accels.mean(axis=0)
            self._comps[1:, :] = self._comps[:-1, :]
            self._comps[0, :] = self._perturb(self._compass, 2.0)
            compass = self._comps.mean(axis=0)
        else:
            gyro = self._gyro
            accel = self._accel
            compass = self._compass
        orient = np.deg2rad(self._orientation)
        self._write(self._read()._replace(
            timestamp=timestamp,
            accel=V(
                int(clamp(accel[0], -8, 8) * ACCEL_FACTOR),
                int(clamp(accel[1], -8, 8) * ACCEL_FACTOR),
                int(clamp(accel[2], -8, 8) * ACCEL_FACTOR),
                ),
            gyro=V(
                int(clamp(gyro[0], -500, 500) * GYRO_FACTOR),
                int(clamp(gyro[1], -500, 500) * GYRO_FACTOR),
                int(clamp(gyro[2], -500, 500) * GYRO_FACTOR),
                ),
            compass=V(
                int(clamp(compass[0], -4, 4) * COMPASS_FACTOR),
                int(clamp(compass[1], -4, 4) * COMPASS_FACTOR),
                int(clamp(compass[2], -4, 4) * COMPASS_FACTOR),
                ),
            orient=V(
                int(clamp(orient[0], -180, 180) * ORIENT_FACTOR),
                int(clamp(orient[1], -180, 180) * ORIENT_FACTOR),
                int(clamp(orient[2], -180, 180) * ORIENT_FACTOR),
                )
            ))




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

"The Raspberry Pi Sense HAT Emulator library"

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )

import sys

from .sense_hat import SenseHat, SenseHat as AstroPi
from .stick import (
    SenseStick,
    InputEvent,
    DIRECTION_UP,
    DIRECTION_DOWN,
    DIRECTION_LEFT,
    DIRECTION_RIGHT,
    DIRECTION_MIDDLE,
    ACTION_PRESSED,
    ACTION_RELEASED,
    ACTION_HELD,
    )

__project__      = 'sense-emu'
__version__      = '1.0'
__author__       = 'Raspberry Pi Foundation'
__author_email__ = 'info@raspberrypi.org'
__url__          = 'http://sense-emu.readthedocs.io/'
__platforms__    = ['ALL']

__classifiers__ = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Environment :: X11 Applications :: GTK',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
    'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
    'Operating System :: POSIX :: Linux',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Topic :: Scientific/Engineering',
    ]

__keywords__ = [
    'raspberrypi', 'sense', 'hat'
    ]

__requires__ = [
    'numpy', 'Pillow',
    ]

__extra_requires__ = {
    'doc':   ['sphinx'],
    'test':  ['pytest', 'coverage', 'mock'],
    }

if sys.version_info[:2] == (3, 2):
    # Particular versions are required for Python 3.2 compatibility
    __extra_requires__['doc'].extend([
        'Jinja2<2.7',
        'MarkupSafe<0.16',
        ])
    __extra_requires__['test'][1] = 'coverage<4.0dev'

__entry_points__ = {
    'console_scripts': [
        'sense_rec = sense_emu.record:app',
        'sense_play = sense_emu.play:app',
        'sense_csv = sense_emu.dump:app',
        ],
    'gui_scripts': [
        'sense_emu_gui = sense_emu.gui:main',
        ],
    }





import numpy as np
from time import sleep
from sense_emu import SenseHat


def clamp(value, min_value, max_value):
    """
    Returns *value* clamped to the range *min_value* to *max_value* inclusive.
    """
    return min(max_value, max(min_value, value))

def scale(value, from_min, from_max, to_min=0, to_max=7):
    """
    Returns *value*, which is expected to be in the range *from_min* to
    *from_max* inclusive, scaled to the range *to_min* to *to_max* inclusive.
    If *value* is not within the expected range, the result is not guaranteed
    to be in the scaled range.
    """
    from_range = from_max - from_min
    to_range = to_max - to_min
    return (((value - from_min) / from_range) * to_range) + to_min

def display_readings(hat):
    """
    Display the temperature, pressure, and humidity readings of the HAT as red,
    green, and blue bars on the screen respectively.
    """
    temperature_range = (0, 40)
    pressure_range = (950, 1050)
    humidity_range = (0, 100)
    temperature = 7 - round(scale(clamp(hat.temperature, *temperature_range), *temperature_range))
    pressure = 7 - round(scale(clamp(hat.pressure, *pressure_range), *pressure_range))
    humidity = 7 - round(scale(clamp(hat.humidity, *humidity_range), *humidity_range))
    # Scroll screen 1 pixel left, clear the right column, and render new points
    screen = np.array(hat.get_pixels(), dtype=np.uint8).reshape((8, 8, 3))
    screen[:, :-1, :] = screen[:, 1:, :]
    screen[:, 7, :] = (0, 0, 0)
    screen[temperature, 7, :] += np.array((255, 0, 0), dtype=np.uint8)
    screen[pressure, 7, :] += np.array((0, 255, 0), dtype=np.uint8)
    screen[humidity, 7, :] += np.array((0, 0, 255), dtype=np.uint8)
    hat.set_pixels([pixel for row in screen for pixel in row])


hat = SenseHat()
hat.clear()
while True:
    display_readings(hat)
    sleep(1)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import os
import logging
import argparse
import datetime as dt
from time import time, sleep
from struct import Struct

from . import __version__
from .i18n import _
from .terminal import TerminalApplication, FileType
from .common import HEADER_REC, DATA_REC, DataRecord
from .imu import IMUServer
from .pressure import PressureServer
from .humidity import HumidityServer
from .lock import EmulatorLock


class PlayApplication(TerminalApplication):
    def __init__(self):
        super(PlayApplication, self).__init__(
            version=__version__,
            description=_("Replays readings recorded from a Raspberry Pi "
                "Sense HAT, via the Sense HAT emulation library."))
        self.parser.add_argument('input', type=FileType('rb'))

    def source(self, f):
        logging.info(_('Reading header'))
        magic, ver, offset = HEADER_REC.unpack(f.read(HEADER_REC.size))
        if magic != b'SENSEHAT':
            raise IOError(_('Invalid magic number at start of input'))
        if ver != 1:
            raise IOError(_('Unrecognized file version number (%d)') % ver)
        logging.info(
            _('Playing back recording taken at %s'),
            dt.datetime.fromtimestamp(offset).strftime('%c'))
        offset = time() - offset
        while True:
            buf = f.read(DATA_REC.size)
            if not buf:
                break
            elif len(buf) < DATA_REC.size:
                raise IOError(_('Incomplete data record at end of file'))
            else:
                data = DataRecord(*DATA_REC.unpack(buf))
                yield data._replace(timestamp=data.timestamp + offset)

    def main(self, args):
        lock = EmulatorLock('sense_play')
        try:
            lock.acquire()
        except:
            logging.error(
                'Another process is currently acting as the Sense HAT '
                'emulator')
            return 1
        try:
            imu = IMUServer(simulate_world=False)
            psensor = PressureServer(simulate_noise=False)
            hsensor = HumidityServer(simulate_noise=False)
            skipped = 0
            for rec, data in enumerate(self.source(args.input)):
                now = time()
                if data.timestamp < now:
                    if not skipped:
                        logging.warning(_('Skipping records to catch up'))
                    skipped += 1
                    continue
                else:
                    sleep(data.timestamp - now)
                psensor.set_values(data.pressure, data.ptemp)
                hsensor.set_values(data.humidity, data.htemp)
                imu.set_imu_values(
                    (data.ax, data.ay, data.az),
                    (data.gx, data.gy, data.gz),
                    (data.cx, data.cy, data.cz),
                    (data.ox, data.oy, data.oz),
                    )
            if skipped:
                logging.warning(_('Skipped %d records during playback'), skipped)
            logging.info(_('Finished playback of %d records'), rec)
        finally:
            lock.release()


app = PlayApplication()



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

"""
Defines base classes for command line utilities.

This module define a TerminalApplication class which provides common facilities
to command line applications: a help screen, universal file globbing, response
file handling, and common logging configuration and options.
"""

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import sys
import io
import os
import argparse
import textwrap
import logging
import locale
import traceback

from . import configparser
from .i18n import init_i18n, _

try:
    # Optionally import argcomplete (for auto-completion) if it's installed
    import argcomplete
except ImportError:
    argcomplete = None


# Set up a console logging handler which just prints messages without any other
# adornments. This will be used for logging messages sent before we "properly"
# configure logging according to the user's preferences
init_i18n()
_CONSOLE = logging.StreamHandler(sys.stderr)
_CONSOLE.setFormatter(logging.Formatter('%(message)s'))
_CONSOLE.setLevel(logging.DEBUG)
logging.getLogger().addHandler(_CONSOLE)


class FileType(object):
    # Variant of argparse.FileType that handles binary stdin/stdout streams
    # correctly under Python 3, and handles encoded text files correctly under
    # Python 2
    def __init__(self, mode='r', bufsize=-1, encoding=None, errors=None):
        self._mode = mode
        self._bufsize = bufsize
        self._encoding = encoding
        self._errors = errors

    def __call__(self, string):
        if string == '-':
            if 'r' in self._mode:
                if 'b' in self._mode:
                    try:
                        return sys.stdin.buffer
                    except AttributeError:
                        pass
                return sys.stdin
            elif 'w' in self._mode:
                if 'b' in self._mode:
                    try:
                        return sys.stdout.buffer
                    except AttributeError:
                        pass
                return sys.stdout
            else:
                raise ValueError(_('argument "-" with mode %r') % self._mode)
        try:
            return io.open(string, self._mode, self._bufsize, self._encoding, self._errors)
        except IOError as e:
            raise argparse.ArgumentTypeError(
                    _("can't open '%(name)s': %(error)s") % {'name': string, 'error': e})

    def __repr__(self):
        args = self._mode, self._bufsize
        kwargs = [('encoding', self._encoding), ('errors', self._errors)]
        args_str = ', '.join([repr(arg) for arg in args if arg != -1] +
                             ['%s=%r' % (kw, arg) for kw, arg in kwargs
                              if arg is not None])
        return '%s(%s)' % (type(self).__name__, args_str)


class TerminalApplication(object):
    """
    Base class for command line applications.

    This class provides command line parsing, file globbing, response file
    handling and common logging configuration for command line utilities.
    Descendent classes should override the main() method to implement their
    main body, and __init__() if they wish to extend the command line options.
    """
    # Get the default output encoding from the default locale
    encoding = locale.getdefaultlocale()[1]

    # This class is the abstract base class for each of the command line
    # utility classes defined. It provides some basic facilities like an option
    # parser, console pretty-printing, logging and exception handling

    def __init__(
            self, version, description=None, config_files=None,
            config_section=None, config_bools=None):
        super(TerminalApplication, self).__init__()
        if description is None:
            description = self.__doc__
        self.parser = argparse.ArgumentParser(
            description=description,
            fromfile_prefix_chars='@')
        self.parser.add_argument(
            '--version', action='version', version=version)
        if config_files:
            self.config = configparser.ConfigParser(interpolation=None)
            self.config_files = config_files
            self.config_section = config_section
            self.config_bools = config_bools
            self.parser.add_argument(
                '-c', '--config', metavar='FILE',
                help=_('specify the configuration file to load'))
        else:
            self.config = None
        self.parser.set_defaults(log_level=logging.WARNING)
        self.parser.add_argument(
            '-q', '--quiet', dest='log_level', action='store_const',
            const=logging.ERROR, help=_('produce less console output'))
        self.parser.add_argument(
            '-v', '--verbose', dest='log_level', action='store_const',
            const=logging.INFO, help=_('produce more console output'))
        opt = self.parser.add_argument(
            '-l', '--log-file', metavar='FILE',
            help=_('log messages to the specified file'))
        if argcomplete:
            # XXX Complete with *.log, *.txt
            #opt.completer = ???
            pass
        self.parser.add_argument(
            '-P', '--pdb', dest='debug', action='store_true', default=False,
            help=_('run under PDB (debug mode)'))

    def __call__(self, args=None):
        if args is None:
            args = sys.argv[1:]
        if argcomplete:
            argcomplete.autocomplete(self.parser, exclude=['-P'])
        elif 'COMP_LINE' in os.environ:
            return 0
        sys.excepthook = self.handle
        args = self.read_configuration(args)
        args = self.parser.parse_args(args)
        self.configure_logging(args)
        if args.debug:
            try:
                import pudb
            except ImportError:
                pudb = None
                import pdb
            return (pudb or pdb).runcall(self.main, args)
        else:
            return self.main(args) or 0

    def read_configuration(self, args):
        if not self.config:
            return args
        # Parse the --config argument only
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument('-c', '--config', dest='config', action='store')
        conf_args, args = parser.parse_known_args(args)
        if conf_args.config:
            self.config_files.append(conf_args.config)
        logging.info(
            _('Reading configuration from %s'), ', '.join(self.config_files))
        conf_read = self.config.read(self.config_files)
        if conf_args.config and conf_args.config not in conf_read:
            self.parser.error('unable to read %s' % conf_args.config)
        if conf_read:
            if self.config_bools is None:
                self.config_bools = ['pdb']
            else:
                self.config_bools = ['pdb'] + self.config_bools
            if not self.config_section:
                self.config_section = self.config.sections()[0]
            if not self.config_section in self.config.sections():
                self.parser.error(
                    _('unable to locate [%s] section in configuration') % self.config_section)
            self.parser.set_defaults(**{
                key:
                self.config.getboolean(self.config_section, key)
                if key in self.config_bools else
                self.config.get(self.config_section, key)
                for key in self.config.options(self.config_section)
                })
        return args

    def configure_logging(self, args):
        _CONSOLE.setLevel(args.log_level)
        if args.log_file:
            log_file = logging.FileHandler(args.log_file)
            log_file.setFormatter(
                logging.Formatter('%(asctime)s, %(levelname)s, %(message)s'))
            log_file.setLevel(logging.DEBUG)
            logging.getLogger().addHandler(log_file)
        if args.debug:
            logging.getLogger().setLevel(logging.DEBUG)
        else:
            logging.getLogger().setLevel(logging.INFO)

    def handle(self, exc_type, exc_value, exc_trace):
        "Global application exception handler"
        if issubclass(exc_type, (SystemExit,)):
            # Exit with 0 ("success") for system exit (as it was intentional)
            return 0
        elif issubclass(exc_type, (KeyboardInterrupt,)):
            # Exit with 2 if the user deliberately terminates with Ctrl+C
            return 2
        elif issubclass(exc_type, (argparse.ArgumentError,)):
            # For option parser errors output the error along with a message
            # indicating how the help page can be displayed
            logging.critical(str(exc_value))
            logging.critical(_('Try the --help option for more information.'))
            return 2
        elif issubclass(exc_type, (IOError,)):
            # For simple errors like IOError just output the message which
            # should be sufficient for the end user (no need to confuse them
            # with a full stack trace)
            logging.critical(str(exc_value))
            return 1
        else:
            # Otherwise, log the stack trace and the exception into the log
            # file for debugging purposes
            for line in traceback.format_exception(exc_type, exc_value, exc_trace):
                for msg in line.rstrip().split('\n'):
                    logging.critical(msg.replace('%', '%%'))
            return 1

    def main(self, args):
        "Called as the main body of the utility"
        raise NotImplementedError





import numpy as np
from time import sleep
from sense_emu import SenseHat


def clamp(value, min_value, max_value):
    """
    Returns *value* clamped to the range *min_value* to *max_value* inclusive.
    """
    return min(max_value, max(min_value, value))

def scale(value, from_min, from_max, to_min=0, to_max=8):
    """
    Returns *value*, which is expected to be in the range *from_min* to
    *from_max* inclusive, scaled to the range *to_min* to *to_max* inclusive.
    If *value* is not within the expected range, the result is not guaranteed
    to be in the scaled range.
    """
    from_range = from_max - from_min
    to_range = to_max - to_min
    return (((value - from_min) / from_range) * to_range) + to_min

def render_bar(screen, origin, width, height, color):
    """
    Fills a rectangle within *screen* based at *origin* (an ``(x, y)`` tuple),
    *width* pixels wide and *height* pixels high. The rectangle will be filled
    in *color*.
    """
    # Calculate the coordinates of the boundaries
    x1, y1 = origin
    x2 = x1 + width
    y2 = y1 + height
    # Invert the Y-coords so we're drawing bottom up
    max_y, max_x = screen.shape[:2]
    y1, y2 = max_y - y2, max_y - y1
    # Draw the bar
    screen[y1:y2, x1:x2, :] = color

def display_readings(hat):
    """
    Display the temperature, pressure, and humidity readings of the HAT as red,
    green, and blue bars on the screen respectively.
    """
    # Calculate the environment values in screen coordinates
    temperature_range = (0, 40)
    pressure_range = (950, 1050)
    humidity_range = (0, 100)
    temperature = scale(clamp(hat.temperature, *temperature_range), *temperature_range)
    pressure = scale(clamp(hat.pressure, *pressure_range), *pressure_range)
    humidity = scale(clamp(hat.humidity, *humidity_range), *humidity_range)
    # Render the bars
    screen = np.zeros((8, 8, 3), dtype=np.uint8)
    render_bar(screen, (0, 0), 2, round(temperature), color=(255, 0, 0))
    render_bar(screen, (3, 0), 2, round(pressure), color=(0, 255, 0))
    render_bar(screen, (6, 0), 2, round(humidity), color=(0, 0, 255))
    hat.set_pixels([pixel for row in screen for pixel in row])


hat = SenseHat()
while True:
    display_readings(hat)
    sleep(0.1)



from sense_emu import SenseHat

sense = SenseHat()

red = (255, 0, 0)
blue = (0, 0, 255)

while True:
    temp = sense.temp
    pixels = [red if i < temp else blue for i in range(64)]
    sense.set_pixels(pixels)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
nstr = str
str = type('')


import mmap
from time import time

import numpy as np

from .pressure import init_pressure, PRESSURE_DATA, PressureData, PRESSURE_FACTOR, TEMP_FACTOR, TEMP_OFFSET
from .humidity import init_humidity, HUMIDITY_DATA, HumidityData
from .imu import init_imu, IMU_DATA, IMUData, ACCEL_FACTOR, GYRO_FACTOR, COMPASS_FACTOR, ORIENT_FACTOR


class Settings(object):
    def __init__(self, path):
        self.path = path


class RTIMU(object):
    def __init__(self, settings):
        self.settings = settings
        self._fd = init_imu()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_READ)
        self._last_data = None
        self._imu_data = {
            'accel':            (0.0, 0.0, 0.0),
            'accelValid':       False,
            'compass':          (0.0, 0.0, 0.0),
            'compassValid':     False,
            'fusionPose':       (0.0, 0.0, 0.0),
            'fusionPoseValid':  False,
            'fusionQPose':      (0.0, 0.0, 0.0, 0.0),
            'fusionQPoseValid': False,
            'gyro':             (0.0, 0.0, 0.0),
            'gyroValid':        False,
            'humidity':         float('nan'),
            'humidityValid':    False,
            'pressure':         float('nan'),
            'pressureValid':    False,
            'temperature':      float('nan'),
            'temperatureValid': False,
            'timestamp':        0,
            }

    def _read(self):
        (
            type, name, timestamp,
            ax, ay, az,
            gx, gy, gz,
            cx, cy, cz,
            ox, oy, oz,
            ) = IMU_DATA.unpack_from(self._map)
        return IMUData(
            type, name, timestamp,
            np.array((ax, ay, az)),
            np.array((gx, gy, gz)),
            np.array((cx, cy, cz)),
            np.array((ox, oy, oz)),
            )

    def IMUInit(self):
        self._last_data = self._read()
        return self._last_data.type != 0

    def IMUGetPollInterval(self):
        return 10 # 3 on the actual board

    def IMUGetGyroBiasValid(self):
        raise NotImplementedError

    def IMURead(self):
        data = self._read()
        if data.timestamp == self._last_data.timestamp:
            return False
        else:
            self._last_data = data
            self._imu_data = {
                'accel':            tuple(data.accel / ACCEL_FACTOR),
                'accelValid':       True,
                'compass':          tuple((data.compass / COMPASS_FACTOR) * 100), # convert Gauss to uT
                'compassValid':     True,
                'fusionPose':       tuple(data.orient / ORIENT_FACTOR),
                'fusionPoseValid':  True,
                'fusionQPose':      (0.0, 0.0, 0.0, 0.0),
                'fusionQPoseValid': False,
                'gyro':             tuple(data.gyro / GYRO_FACTOR),
                'gyroValid':        True,
                'humidity':         float('nan'),
                'humidityValid':    False,
                'pressure':         float('nan'),
                'pressureValid':    False,
                'temperature':      float('nan'),
                'temperatureValid': False,
                'timestamp':        data.timestamp,
                }
            return True

    def IMUType(self):
        return self._read().type # 6 in real unit

    def IMUName(self):
        return self._read().name.decode('ascii') # "LSM9DS1" in real unit

    def getAccel(self):
        return self._imu_data['accel']

    def getAccelCalibrationValid(self):
        raise NotImplementedError

    def getAccelResiduals(self):
        raise NotImplementedError

    def getCompass(self):
        return self._imu_data['compass']

    def getCompassCalibrationEllipsoidValid(self):
        raise NotImplementedError

    def getCompassCalibrationValid(self):
        raise NotImplementedError

    def getFusionData(self):
        return self._imu_data['fusionPose']

    def getGyro(self):
        return self._imu_data['gyro']

    def getIMUData(self):
        return self._imu_data

    def getMeasuredPose(self):
        raise NotImplementedError

    def getMeasuredQPose(self):
        raise NotImplementedError

    def setCompassEnable(self, value):
        pass

    def setGyroEnable(self, value):
        pass

    def setAccelEnable(self, value):
        pass


class RTPressure(object):
    def __init__(self, settings):
        self.settings = settings
        self._fd = init_pressure()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_READ)
        self._last_read = 0.0
        self._last_data = None
        self._p_ref = None

    def _read(self):
        now = time()
        if now - self._last_read > 0.04:
            self._last_read = now
            self._last_data = PressureData(*PRESSURE_DATA.unpack_from(self._map))
        return self._last_data

    def pressureInit(self):
        d = self._read()
        self._p_ref = d.P_REF
        return d.type != 0

    def pressureRead(self):
        if self._p_ref is None:
            return (0, 0.0, 0, 0.0)
        else:
            d = self._read()
            return (
                d.P_VALID, d.P_OUT / PRESSURE_FACTOR,
                d.T_VALID, d.T_OUT / TEMP_FACTOR + TEMP_OFFSET,
                )

    def pressureType(self):
        return self._read().type

    def pressureName(self):
        return self._read().name.decode('ascii')


class RTHumidity(object):
    def __init__(self, settings):
        self.settings = settings
        self._fd = init_humidity()
        self._map = mmap.mmap(self._fd.fileno(), 0, access=mmap.ACCESS_READ)
        self._last_read = 0.0
        self._last_data = None
        self._humidity_m = None
        self._humidity_c = None
        self._temp_m = None
        self._temp_c = None

    def _read(self):
        now = time()
        if now - self._last_read > 0.13:
            self._last_read = now
            self._last_data = HumidityData(*HUMIDITY_DATA.unpack_from(self._map))
        return self._last_data

    def humidityInit(self):
        d = self._read()
        try:
            self._humidity_m = (d.H1 - d.H0) / (d.H1_OUT - d.H0_OUT)
            self._humidity_c = d.H0 - self._humidity_m * d.H0_OUT
            self._temp_m = (d.T1 - d.T0) / (d.T1_OUT - d.T0_OUT)
            self._temp_c = d.T0 - self._temp_m * d.T0_OUT
            return True
        except ZeroDivisionError:
            return False

    def humidityRead(self):
        if self._temp_m is None:
            return (0, 0.0, 0, 0.0)
        else:
            d = self._read()
            return (
                d.H_VALID, d.H_OUT * self._humidity_m + self._humidity_c,
                d.T_VALID, d.T_OUT * self._temp_m + self._temp_c,
                )

    def humidityType(self):
        return self._read().type

    def humidityName(self):
        return self._read().name.decode('ascii')





#!/usr/bin/python3
from sense_emu import SenseHat

x = y = 4
hat = SenseHat()

def update_screen():
    hat.clear()
    hat.set_pixel(x, y, 255, 255, 255)

def clamp(value, min_value=0, max_value=7):
    return min(max_value, max(min_value, value))

def move_dot(event):
    global x, y
    if event.action in ('pressed', 'held'):
        x = clamp(x + {
            'left': -1,
            'right': 1,
            }.get(event.direction, 0))
        y = clamp(y + {
            'up': -1,
            'down': 1,
            }.get(event.direction, 0))

update_screen()
while True:
    for event in hat.stick.get_events():
        move_dot(event)
        update_screen()



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')


import sys
import os
import csv
import logging
import argparse
import datetime as dt
from time import time, sleep
from struct import Struct

from . import __version__
from .i18n import _
from .terminal import TerminalApplication, FileType
from .common import HEADER_REC, DATA_REC, DataRecord


class DumpApplication(TerminalApplication):
    def __init__(self):
        super(DumpApplication, self).__init__(
            version=__version__,
            description=_("Converts a Sense HAT recording to CSV format, for "
                "the purposes of debugging or analysis."))
        self.parser.add_argument(
            '--timestamp-format', action='store', default='%Y-%m-%dT%H:%M:%S.%f', metavar='FMT',
            help=_('the format to use when outputting the record timestamp '
            '(default: %(default)s)'))
        self.parser.add_argument(
            '--header', action='store_true', default=False,
            help=_('if specified, output column headers'))
        self.parser.add_argument('input', type=FileType('rb'))
        # Eurgh ... csv module under Python 2 only outputs byte-strings
        if sys.version_info.major == 2:
            self.parser.add_argument('output', type=FileType('wb'))
        else:
            self.parser.add_argument('output', type=FileType('w', encoding='utf-8'))

    def source(self, f):
        logging.info(_('Reading header'))
        magic, ver, offset = HEADER_REC.unpack(f.read(HEADER_REC.size))
        if magic != b'SENSEHAT':
            raise IOError(_('Invalid magic number at start of input'))
        if ver != 1:
            raise IOError(_('Unrecognized file version number (%d)') % ver)
        logging.info(
            _('Dumping recording taken at %s'),
            dt.datetime.fromtimestamp(offset).strftime('%c'))
        offset = time() - offset
        while True:
            buf = f.read(DATA_REC.size)
            if not buf:
                break
            elif len(buf) < DATA_REC.size:
                raise IOError(_('Incomplete data record at end of file'))
            else:
                yield DataRecord(*DATA_REC.unpack(buf))

    def main(self, args):
        writer = csv.writer(args.output)
        if args.header:
            writer.writerow((
                'timestamp',
                'pressure', 'pressure_temp',
                'humidity', 'humidity_temp',
                'accel_x', 'accel_y', 'accel_z',
                'gyro_x', 'gyro_y', 'gyro_z',
                'compass_x', 'compass_y', 'compass_z',
                'orient_x', 'orient_y', 'orient_z',
                ))
        for rec, data in enumerate(self.source(args.input)):
            writer.writerow((
                dt.datetime.fromtimestamp(data.timestamp).strftime(args.timestamp_format),
                data.pressure, data.ptemp,
                data.humidity, data.htemp,
                data.ax, data.ay, data.az,
                data.gx, data.gy, data.gz,
                data.cx, data.cy, data.cz,
                data.ox, data.oy, data.oz,
                ))
        logging.info(_('Converted %d records'), rec)


app = DumpApplication()



#!/usr/bin/env python3

# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# All Rights Reserved.

"""An emulator for the Raspberry Pi Sense HAT."""

import os
import sys
from setuptools import setup, find_packages

if sys.version_info[0] == 2:
    if not sys.version_info >= (2, 7):
        raise ValueError('This package requires Python 2.7 or newer')
elif sys.version_info[0] == 3:
    if not sys.version_info >= (3, 2):
        raise ValueError('This package requires Python 3.2 or newer')
else:
    raise ValueError('Unrecognized major version of Python')

HERE = os.path.abspath(os.path.dirname(__file__))

# Workaround <http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html>
try:
    import multiprocessing
except ImportError:
    pass

# Mock out dependency modules while installing
class Mock(object):
    __all__ = []

    def __init__(self, *args, **kw):
        pass

    def __call__(self, *args, **kw):
        return Mock()

    def __mul__(self, other):
        return Mock()

    def __and__(self, other):
        return Mock()

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False

    @classmethod
    def __getattr__(cls, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        else:
            return Mock()

def main():
    sys.modules['numpy'] = Mock()
    sys.modules['RTIMU'] = Mock()
    sys.modules['PIL'] = Mock()
    import io
    import sense_emu as app
    with io.open(os.path.join(HERE, 'README.rst'), 'r') as readme:
        setup(
            name                 = app.__project__,
            version              = app.__version__,
            description          = app.__doc__,
            long_description     = readme.read(),
            classifiers          = app.__classifiers__,
            author               = app.__author__,
            author_email         = app.__author_email__,
            url                  = app.__url__,
            license              = [
                c.rsplit('::', 1)[1].strip()
                for c in app.__classifiers__
                if c.startswith('License ::')
                ][0],
            keywords             = app.__keywords__,
            packages             = find_packages(),
            include_package_data = True,
            platforms            = app.__platforms__,
            install_requires     = app.__requires__,
            extras_require       = app.__extra_requires__,
            entry_points         = app.__entry_points__,
            )


if __name__ == '__main__':
    main()




# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 2.1 of the License, or (at your option)
# any later version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import sys
import os
import io
import errno
from time import time, sleep


if sys.platform.startswith('win'):
    import ctypes
    kernel32 = ctypes.windll.kernel32
    DWORD = ctypes.c_ulong
    PROCESS_QUERY_INFORMATION = 0x0400
    ERROR_ACCESS_DENIED = 0x5
    ERROR_INVALID_PARAMETER = 0x57
    STILL_ACTIVE = 259

    def pid_exists(pid):
        h = kernel32.OpenProcess(PROCESS_QUERY_INFORMATION, 0, pid)
        try:
            if not h:
                if kernel32.GetLastError() == ERROR_ACCESS_DENIED:
                    # If access is denied, there's obviously a process denying
                    # access...
                    return True
                elif kernel32.GetLastError() == ERROR_INVALID_PARAMETER:
                    # Invalid parameter is no such process
                    return False
                raise OSError('unable to get handle for pid %d' % pid)
            out = DWORD()
            if kernel32.GetExitCodeProcess(h, ctypes.byref(out)):
                return out.value == STILL_ACTIVE
            raise OSError('unable to query exit code for pid %d' % pid)
        finally:
            kernel32.CloseHandle(h)
else:
    def pid_exists(pid):
        if pid == 0:
            return True
        try:
            os.kill(pid, 0)
        except OSError as e:
            if e.errno == errno.ESRCH:
                return False
            elif e.errno == errno.EPERM:
                return True
            else:
                raise
        else:
            return True


def lock_filename():
    """
    Return the filename used as a lock-file by applications that can drive the
    emulation (currently sense_emu_gui and sense_play). On UNIX we try
    ``/dev/shm`` then fall back to ``/tmp``; on Windows we use whatever
    ``%TEMP%`` contains
    """
    fname = 'rpi-sense-emu-pid'
    if sys.platform.startswith('win'):
        # just use a temporary file on Windows
        return os.path.join(os.environ['TEMP'], fname)
    else:
        if os.path.exists('/dev/shm'):
            return os.path.join('/dev/shm', fname)
        else:
            return os.path.join('/tmp', fname)


class EmulatorLock(object):
    def __init__(self, name):
        self._filename = lock_filename()
        self.name = name # XXX not currently used

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.release()

    def acquire(self, timeout=None):
        """
        Acquire the emulator lock. This is expected to be called by anything
        wishing to drive the emulator's registers (sense_emu_gui and sense_play
        currently).
        """
        if self._is_stale():
            self._break_lock()
        self._write_pid()

    def release(self):
        """
        Release the emulator lock (presumably after :meth:`acquire`).
        """
        self._break_lock()

    def wait(self, timeout=None):
        """
        Wait for a process to acquire the lock. This is expected to be called
        by anything wishing to read the registers and wanting to ensure there's
        something driving them (i.e. any consumer of SenseHat).

        Returns ``True`` if the lock was acquired before *timeout* seconds
        elapsed, or ``False`` otherwise. If *timeout* is ``None`` (the default)
        wait indefinitely.
        """
        end = time()
        if timeout is None:
            wait = 0.1
        else:
            end += timeout
            wait = max(0, timeout / 10)
        while not self._is_held() or self._is_stale():
            if time() > end:
                return False
            sleep(wait)
        return True

    @property
    def mine(self):
        """
        Returns True if the current process holds the lock.
        """
        return self._read_pid() == os.getpid()

    def _is_held(self):
        return os.path.exists(self._filename)

    def _is_stale(self):
        # True if the lock file exists, but the PID it references doesn't
        pid = self._read_pid()
        if pid is not None:
            return not pid_exists(pid)
        else:
            return False

    def _break_lock(self):
        # Unconditionally delete the file
        try:
            os.unlink(self._filename)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    def _read_pid(self):
        try:
            lockfile = io.open(self._filename, 'rb')
        except IOError:
            return None
        else:
            try:
                return int(lockfile.readline().decode('ascii').strip())
            except ValueError:
                return None
            finally:
                lockfile.close()

    def _write_pid(self):
        try:
            lockfile = io.open(self._filename, 'x')
        except ValueError as e:
            # We're on py2.x or 3.2
            mode = 0o666
            lockfile = os.fdopen(os.open(
                self._filename, os.O_CREAT | os.O_EXCL | os.O_WRONLY, mode), 'w')
        lockfile.write('%d\n' % os.getpid())
        lockfile.close()




from time import sleep
from sense_emu import SenseHat
from PIL import Image, ImageDraw

hat = SenseHat()

hat.clear()
origin = (7, 7)
while True:
    a = hat.get_accelerometer_raw()
    # Use the old trick of drawing something too big then down-sizing to get an
    # anti-aliased line
    img = Image.new('RGB', (15, 15))
    draw = ImageDraw.Draw(img)
    dest = (origin[0] + a['x'] * 7.0, origin[1] + a['y'] * 7.0)
    draw.line([origin, dest], fill=(255, 255, 255), width=3)
    img = img.resize((8, 8), Image.BILINEAR)
    hat.set_pixels(list(img.getdata()))
    sleep(0.04)



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# Raspberry Pi Sense HAT Emulator library for the Raspberry Pi
# Copyright (c) 2016 Raspberry Pi Foundation <info@raspberrypi.org>
#
# This package is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This package is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>


from __future__ import (
    unicode_literals,
    absolute_import,
    print_function,
    division,
    )
str = type('')

import os
import logging
import argparse
from threading import Thread, Event
from time import time, sleep
from struct import Struct

from . import __version__
from .i18n import _
from .terminal import TerminalApplication, FileType
from .common import HEADER_REC, DATA_REC


class RecordApplication(TerminalApplication):
    def __init__(self):
        super(RecordApplication, self).__init__(
            version=__version__,
            description=_("Records the readings from a Raspberry Pi Sense HAT "
                "in real time, outputting the results to the specified file."))
        self.parser.add_argument(
            '-c', '--config', dest='config', action='store',
            default='/etc/RTIMULib.ini', metavar='FILE',
            help=_('the Sense HAT configuration file to use (default: %(default)s)'))
        self.parser.add_argument(
            '-d', '--duration', dest='duration', action='store', default=0.0,
            type=float, metavar='SECS',
            help=_('the duration to record for in seconds (default: record '
            'until terminated with Ctrl+C)'))
        self.parser.add_argument(
            '-f', '--flush', dest='flush', action='store_true', default=False,
            help=_('flush every record to disk immediately; reduces chances of '
            'truncated data on power loss, but greatly increases disk activity'))
        self.parser.add_argument('output', type=FileType('wb'))

    def main(self, args):
        try:
            import RTIMU
        except ImportError:
            raise IOError(
                _('unable to import RTIMU; ensure the Sense HAT library is '
                'correctly installed'))
        if not args.config.endswith('.ini'):
            raise argparse.ArgumentError(_('configuration filename must end with .ini'))

        logging.info(_('Reading settings from %s'), args.config)
        settings = RTIMU.Settings(args.config[:-4])
        logging.info(_('Initializing sensors'))
        imu = RTIMU.RTIMU(settings)
        if not imu.IMUInit():
            raise IOError(_('Failed to initialize Sense HAT IMU'))
        psensor = RTIMU.RTPressure(settings)
        if not psensor.pressureInit():
            raise IOError(_('Failed to initialize Sense HAT pressure sensor'))
        hsensor = RTIMU.RTHumidity(settings)
        if not hsensor.humidityInit():
            raise IOError(_('Failed to initialize Sense HAT humidity sensor'))
        interval = imu.IMUGetPollInterval() / 1000.0 # seconds
        nan = float('nan')

        logging.info(_('Starting recording'))
        rec_count = 0
        if args.duration:
            terminate_at = time() + args.duration
        else:
            terminate_at = time() + 1e100
        args.output.write(HEADER_REC.pack(b'SENSEHAT', 1, time()))
        status_stop = Event()
        def status():
            while not status_stop.wait(1.0):
                logging.info(_('%d records written'), rec_count)
        status_thread = Thread(target=status)
        status_thread.daemon = True
        status_thread.start()
        try:
            while True:
                timestamp = time()
                if imu.IMURead():
                    ax, ay, az = imu.getAccel()
                    gx, gy, gz = imu.getGyro()
                    cx, cy, cz = imu.getCompass()
                    ox, oy, oz = imu.getFusionData()
                    pvalid, pressure, ptvalid, ptemp = psensor.pressureRead()
                    hvalid, humidity, htvalid, htemp = hsensor.humidityRead()
                    args.output.write(DATA_REC.pack(
                        timestamp,
                        pressure if pvalid else nan,
                        ptemp if ptvalid else nan,
                        humidity if hvalid else nan,
                        htemp if htvalid else nan,
                        ax, ay, az,
                        gx, gy, gz,
                        cx, cy, cz,
                        ox, oy, oz,
                        ))
                    if args.flush:
                        args.output.flush()
                    rec_count += 1
                if timestamp > terminate_at:
                    break
                delay = max(0.0, timestamp + interval - time())
                if delay:
                    sleep(delay)
        finally:
            status_stop.set()
            status_thread.join()
            logging.info(_('Finishing recording after %d records'), rec_count)
            args.output.close()


app = RecordApplication()
